import copy
import math
from collections import OrderedDict

import lingerext
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.utils.rnn import PackedSequence
from torch.onnx import is_in_onnx_export

from ..config import config
from ..quant import (normalize_bias_with_config, normalize_data_with_config,
                     normalize_weight_with_config)
from ..utils import (Dump, PlatFormQuant, QuantMode, ScalerBuffer, _slice,
                     _unbind, _unbind_packed, get_max_value, hx_slice)
from .iqtensor import (IQTensor, Quant2IQTensor, from_torch_tensor,
                       platform_to_string, quantlinear)
from .linger_functional import (iqCat, torch_pack_padded_sequence,
                                torch_pad_packed_sequence)
from .ops import ModuleIntConfig
from .requant import Requant

iqcat_sym = iqCat.symbolic


def castor_luna_sigmoid(x_int, scale_x=2048.0):
    l_scale = 11 - int(math.log2(scale_x))

    if l_scale > 0:
        x_int = x_int * pow(2, l_scale)
    else:
        x_int = (x_int * pow(2, l_scale) + 0.5).floor().int()
    y_int = lingerext.luna_iqsigmoid(x_int.contiguous(), float(scale_x))
    y_int.clamp_(0, 2**7-1)
    return y_int


def castor_luna_tanh(x_int, scale_x=2048.0):
    l_scale = 11 - int(math.log2(scale_x))

    if l_scale > 0:
        x_int = x_int * pow(2, l_scale)
    else:
        x_int = (x_int * pow(2, l_scale) + 0.5).floor().int()
    y_int = lingerext.luna_iqtanh(x_int.contiguous(), float(scale_x))
    y_int.clamp_(-2**7, 2**7-1)
    return y_int


def castor_luna_requant(x_int, scale_x=2048.0):
    l_scale = 11 - int(math.log2(scale_x))

    if l_scale > 0:
        x_int = x_int * pow(2, l_scale)
    else:
        x_int = (x_int * pow(2, l_scale) + 0.5).floor().int()
    return x_int


class GRUCellFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input, hidden, weight_ih, weight_hh, bias_ih, bias_hh,
                data_bits, parameter_bits, o_bits,
                running_i, running_h, running_iw, running_hw, running_io, running_ho, running_o,
                scale_i, scale_h, scale_iw, scale_hw, scale_io, scale_ho, scale_o,
                momentum, training, prefix, dump, path, mode, quant,
                is_not_from_iqtensor, clamp_data):
        if training:
            ctx.clamp_data = clamp_data
            ctx.o_bits = o_bits
            save_tensors = [input, hidden, weight_ih,
                            weight_hh, bias_ih, bias_hh]
            q_input, scale_i, max_value_ix = quant.quant(
                input.data, data_bits, mode=mode, quant_data='input')
            q_iweight, scale_iw, max_value_iw = quant.quant(
                weight_ih, parameter_bits, mode=mode, quant_data='weight')
            running_iw.mul_(1-momentum).add_(momentum*max_value_iw)

            q_hidden, scale_h, max_value_hx = quant.quant(
                hidden, data_bits, mode=mode, quant_data='input')
            q_hweight, scale_hw, max_value_hw = quant.quant(
                weight_hh, parameter_bits, mode=mode, quant_data='weight')
            running_hw.mul_(1-momentum).add_(momentum*max_value_hw)

            q_input = q_input.float() if data_bits + parameter_bits <= 16 else q_input.double()
            q_iweight = q_iweight.float() if data_bits + \
                parameter_bits <= 16 else q_iweight.double()
            q_hidden = q_hidden.float() if data_bits + \
                parameter_bits <= 16 else q_hidden.double()
            q_hweight = q_hweight.float() if data_bits + \
                parameter_bits <= 16 else q_hweight.double()

            q_gi_outputs = F.linear(q_input, q_iweight)
            q_gh_outputs = F.linear(q_hidden, q_hweight)

            if bias_ih is not None:
                if config.PlatFormQuant.platform_quant == PlatFormQuant.luna_quant:
                    q_ibias = (bias_ih * scale_iw *
                               scale_i + 0.5).floor().int()
                    if data_bits + parameter_bits <= 16:
                        q_ibias = q_ibias.float()
                    else:
                        q_ibias = q_ibias.double()
                else:
                    assert False, "linger only support luna quant."
                # q_ibias = q_ibias.int()
                q_gi_outputs += q_ibias.view(-1)

                if config.PlatFormQuant.platform_quant == PlatFormQuant.luna_quant:
                    q_hbias = (bias_hh * scale_hw *
                               scale_h + 0.5).floor().int()
                    if data_bits + parameter_bits <= 16:
                        q_hbias = q_hbias.float()
                    else:
                        q_hbias = q_hbias.double()
                else:
                    assert False, "linger only support luna quant."
                q_gh_outputs += q_hbias.view(-1)

            if config.PlatFormQuant.platform_quant == PlatFormQuant.luna_quant:  # QX+QW -> Q11
                l_scale_gi = 11 - int(math.log2(scale_i()*scale_iw()))
                if l_scale_gi > 0:
                    gi = q_gi_outputs * pow(2, l_scale_gi)
                else:
                    gi = (q_gi_outputs * pow(2, l_scale_gi) + 0.5).floor().int()

                l_scale_gh = 11 - int(math.log2(scale_h()*scale_hw()))
                if l_scale_gh > 0:
                    gh = q_gh_outputs * pow(2, l_scale_gh)
                else:
                    gh = (q_gh_outputs * pow(2, l_scale_gh) + 0.5).floor().int()
            else:  # QX+QW -> Q10
                assert False, "linger only support luna quant."

            for_backward_gi = quant.dequant(q_gi_outputs, scale_i*scale_iw)
            for_backward_gh = quant.dequant(q_gh_outputs, scale_h*scale_hw)
            save_tensors += [for_backward_gi, for_backward_gh]

            i_r, i_i, i_n = gi.chunk(3, 1)
            h_r, h_i, h_n = gh.chunk(3, 1)
            if config.PlatFormQuant.platform_quant == PlatFormQuant.luna_quant:
                resetgate = castor_luna_sigmoid(i_r + h_r)  # Q11->Q7
                inputgate = castor_luna_sigmoid(i_i + h_i)  # Q11->Q7
                resetgate_h_n = resetgate * h_n  # Q7*Q11->Q18
                resetgate_h_n = castor_luna_requant(
                    resetgate_h_n, 2**18)  # Q18->Q11
                newgate = castor_luna_tanh(i_n + resetgate_h_n)  # Q11 ->Q7
                new_hidden = (hidden*(2**7) + 0.5).floor().int()
                # Q7*Q7+Q7*(Q7-Q7)->Q14
                hy = newgate * (2**7) + inputgate * (new_hidden - newgate)
                hy = quant.dequant(hy, float(2**14))
            else:
                assert False, "linger only support luna quant."

            save_tensors += [hy.clone()]

            if o_bits is not None:
                hy = normalize_data_with_config(hy, clamp_data)
                q_hy_outputs, scale_o, max_value_o = quant.quant(
                    hy, o_bits, mode=mode, quant_data='output')
                running_o.mul_(1-momentum).add_(momentum*max_value_o)
                running_h.mul_(1-momentum).add_(momentum*max_value_o)
                hy = quant.dequant(q_hy_outputs, scale_o)
            else:  # 为None 时  也得保证  running_h的实际值为running_o
                _, _, fake_max_value_o = quant.quant(
                    hy, 8, mode=mode, quant_data='output')
                running_h.mul_(1-momentum).add_(momentum*fake_max_value_o)

            ctx.save_for_backward(*save_tensors)
        else:
            assert running_i > 0, 'invalid running_i <= 0, please finetune training'
            if weight_ih.dtype == torch.float32:
                if is_not_from_iqtensor:
                    scale_i = ScalerBuffer(quant.running_to_scale(
                        running_i, data_bits, mode=mode))
                scale_h = ScalerBuffer(quant.running_to_scale(
                    running_h, data_bits, mode=mode))
                scale_iw = ScalerBuffer(quant.running_to_scale(
                    running_iw, parameter_bits, mode=mode))
                scale_hw = ScalerBuffer(quant.running_to_scale(
                    running_hw, parameter_bits, mode=mode))

            q_input, scale_i, _ = quant.quant(
                input.data, data_bits, scale_i, mode=mode, quant_data='input')
            q_hidden, scale_h, _ = quant.quant(
                hidden, data_bits, scale_h, mode=mode, quant_data='input')
            q_iweight = None
            q_hweight = None
            if weight_ih.dtype == torch.float32:
                q_iweight, _, _ = quant.quant(
                    weight_ih, parameter_bits, scale_iw, mode=mode, quant_data='weight')
                q_hweight, _, _ = quant.quant(
                    weight_hh, parameter_bits, scale_hw, mode=mode, quant_data='weight')
            else:
                q_iweight = weight_ih.double()
                q_hweight = weight_hh.double()

            q_input = q_input.double()
            q_iweight = q_iweight.double()
            q_hidden = q_hidden.double()
            q_hweight = q_hweight.double()
            q_gi_outputs = F.linear(q_input, q_iweight)
            q_gh_outputs = F.linear(q_hidden, q_hweight)

            if bias_ih is not None:
                if bias_ih.dtype == torch.float32:
                    if config.PlatFormQuant.platform_quant == PlatFormQuant.luna_quant:
                        q_ibias = (bias_ih * scale_iw *
                                   scale_i + 0.5).floor().int()
                        q_hbias = (bias_hh * scale_hw *
                                   scale_h + 0.5).floor().int()
                        if data_bits + parameter_bits <= 16:
                            q_ibias = q_ibias.float().double()
                            q_hbias = q_hbias.float().double()
                        else:
                            q_ibias = q_ibias.double()
                            q_hbias = q_hbias.double()
                    else:
                        assert False, "linger only support luna quant."

                else:
                    q_ibias = bias_ih.double()
                    q_hbias = bias_hh.double()
                q_gi_outputs += q_ibias.view(-1)
                q_gh_outputs += q_hbias.view(-1)

            if config.PlatFormQuant.platform_quant == PlatFormQuant.luna_quant:  # QX+QW -> Q11
                l_scale_gi = 11 - int(math.log2(scale_i()*scale_iw()))
                if l_scale_gi > 0:
                    gi = q_gi_outputs * pow(2, l_scale_gi)
                else:
                    gi = (q_gi_outputs * pow(2, l_scale_gi) + 0.5).floor().int()

                l_scale_gh = 11 - int(math.log2(scale_h()*scale_hw()))
                if l_scale_gh > 0:
                    gh = q_gh_outputs * pow(2, l_scale_gh)
                else:
                    gh = (q_gh_outputs * pow(2, l_scale_gh) + 0.5).floor().int()
            else:  # QX+QW -> Q10
                assert False, "linger only support luna quant."

            i_r, i_i, i_n = gi.chunk(3, 1)
            h_r, h_i, h_n = gh.chunk(3, 1)

            if config.PlatFormQuant.platform_quant == PlatFormQuant.luna_quant:
                resetgate = castor_luna_sigmoid(i_r + h_r)  # Q11->Q7
                inputgate = castor_luna_sigmoid(i_i + h_i)  # Q11->Q7
                resetgate_h_n = resetgate * h_n  # Q7*Q11->Q18
                resetgate_h_n = castor_luna_requant(
                    resetgate_h_n, 2**18)  # Q18->Q11
                newgate = castor_luna_tanh(i_n + resetgate_h_n)  # Q11 ->Q7
                new_hidden = (hidden*(2**7) + 0.5).floor().int()
                # Q7*Q7+Q7*(Q7-Q7)->Q14
                hy = newgate * (2**7) + inputgate * (new_hidden - newgate)
                hy = quant.dequant(hy, float(2**14))
            else:
                assert False, "linger only support luna quant."

            if o_bits is not None:
                assert running_o > 0, 'invalid running_o <=0, please finetune first'
                if weight_ih.dtype == torch.float32:
                    scale_o = ScalerBuffer(quant.running_to_scale(
                        running_o, o_bits, mode=mode))
                q_hy_outputs, _, _ = quant.quant(
                    hy, o_bits, scale_o, mode=mode, quant_data='output')
                hy = quant.dequant(q_hy_outputs, scale_o)

            if dump:
                if bias_ih is not None and bias_hh is not None:
                    name_list = ["input", "hidden", "q_input", "q_hidden", "gi", "gh", "q_gi_outputs", "q_gh_outputs", "q_ibias", "q_hbias", "runnung_i", "running_h", "running_iw",
                                 "running_ih", "running_io", "running_ho", "scale_i", "scale_iw", "scale_h", "scale_hw", "resetgate", "inputgate", "newgate", "output", "q_hy_outputs"]
                    attr_list = [input, hidden, q_input, q_hidden, gi, gh, q_gi_outputs, q_gh_outputs, q_ibias, q_hbias, running_i, running_h, running_iw,
                                 running_hw, running_io, running_ho, scale_i, scale_iw, scale_h, scale_hw, resetgate, inputgate, newgate, hy, q_hy_outputs]
                    Dump.dump_file(prefix, ".GruInt.", zip(
                        name_list, attr_list), path)
                else:
                    name_list = ["input", "hidden", "q_input", "q_hidden", "gi", "gh", "q_gi_outputs", "q_gh_outputs", "runnung_i", "running_h", "running_iw",
                                 "running_ih", "running_io", "running_ho", "scale_i", "scale_iw", "scale_h", "scale_hw", "resetgate", "inputgate", "newgate", "output", "q_hy_outputs"]
                    attr_list = [input, hidden, q_input, q_hidden, gi, gh, q_gi_outputs, q_gh_outputs, running_i, running_h, running_iw,
                                 running_hw, running_io, running_ho, scale_i, scale_iw, scale_h, scale_hw, resetgate, inputgate, newgate, hy, q_hy_outputs]
                    Dump.dump_file(prefix, ".GruInt.", zip(
                        name_list, attr_list), path)

        return hy

    @staticmethod
    def backward(ctx, gradOutput):
        input, hidden, weight_ih, weight_hh, input_bias, hidden_bias, input_gates, hidden_gates, hy = ctx.saved_tensors
        clamp_data = ctx.clamp_data
        o_bits = ctx.o_bits
        hy.requires_grad_(True)
        hidden.requires_grad_(True)
        with torch.enable_grad():
            if o_bits is not None:
                clamp_hy = normalize_data_with_config(hy, clamp_data)
            else:
                clamp_hy = hy
            gradOutput, = torch.autograd.grad(clamp_hy, hy, gradOutput)
        grad_input_gates, grad_hidden_gates, grad_hx, grad_input_bias, grad_hidden_bias = lingerext.gru_cell_backward(
            gradOutput, input_gates, hidden_gates, hidden, input_bias, hidden_bias)
        grad_in = grad_input_gates.matmul(weight_ih)
        grad_w_ih = grad_input_gates.t().matmul(input)
        grad_h_ih = grad_hidden_gates.t().matmul(hidden)

        return grad_in, grad_hx, grad_w_ih, grad_h_ih, grad_input_bias, grad_hidden_bias, None, None, None, None, None, \
            None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, \
            None, None, None, None, None, None, None, None, None, None


class GRUSingleONNXFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input, lengths, hidden_state, weight_ih, weight_hh, bias_ih, bias_hh,
                weight_ih_reverse, weight_hh_reverse, bias_ih_reverse, bias_hh_reverse,
                input_size, hidden_size, num_layers, batch_first, dropout, bidirectional,
                data_bits, parameter_bits, o_bits,
                scale_i, scale_iw, scale_io, scale_h, scale_hw, scale_ho, scale_o,
                scale_i_reverse, scale_iw_reverse, scale_io_reverse, scale_h_reverse, scale_hw_reverse, scale_ho_reverse, scale_o_reverse,
                is_not_from_iqtensor):
        output = None
        hidden_state = None
        batch_size = None
        seq_length = None
        num_directions = 2 if bidirectional else 1
        if batch_first:
            batch_size = input.size(0)
            seq_length = input.size(
                1) if lengths is None else torch.max(lengths)
            output = torch.randn(batch_size, seq_length,
                                 hidden_size*num_directions, device=input.device)
        else:
            batch_size = input.size(1)
            seq_length = input.size(
                0) if lengths is None else torch.max(lengths)
            output = torch.randn(seq_length, batch_size,
                                 hidden_size*num_directions, device=input.device)
        hidden_state = torch.zeros(
            num_directions, batch_size, hidden_size, device=input.device)
        return output, hidden_state

    @staticmethod
    def backward(ctx, gradOutput, gradHidden):

        return None, None, None, None, None, None,\
            None, None, None, None,\
            None, None, None, None, None, None,\
            None, None, None,\
            None, None, None, None, None, None, None,\
            None, None, None, None, None, None, None,\
            None, None, None, None

    @staticmethod
    def symbolic(g, input, lengths, hidden_state, weight_ih, weight_hh, bias_ih, bias_hh,
                 weight_ih_reverse, weight_hh_reverse, bias_ih_reverse, bias_hh_reverse,
                 input_size, hidden_size, num_layers, batch_first, dropout, bidirectional,
                 data_bits, parameter_bits, o_bits,
                 scale_i, scale_iw, scale_io, scale_h, scale_hw, scale_ho, scale_o,
                 scale_i_reverse, scale_iw_reverse, scale_io_reverse, scale_h_reverse, scale_hw_reverse, scale_ho_reverse, scale_o_reverse,
                 is_not_from_iqtensor):

        param_dict = {'input_size_i': input_size, 'hidden_size_i': hidden_size, 'num_layers_i': num_layers,
                      'batch_first_i': batch_first, 'dropout_f': 0, 'go_forward_i': True,
                      'scale_i_f': scale_i(), 'scale_h_f': scale_h(), 'scale_iw_f': scale_iw(), 'scale_hw_f': scale_hw(),
                      'data_bits_i': data_bits, 'parameter_bits_i': parameter_bits,
                      }
        param_back_dict = {}
        if bidirectional:
            param_back_dict = {'input_size_i': input_size, 'hidden_size_i': hidden_size, 'num_layers_i': num_layers,
                               'batch_first_i': batch_first, 'dropout_f': 0, 'go_forward_i': False,
                               'scale_i_f': scale_i_reverse(), 'scale_h_f': scale_h_reverse(), 'scale_iw_f': scale_iw_reverse(), 'scale_hw_f': scale_hw_reverse(),
                               'data_bits_i': data_bits, 'parameter_bits_i': parameter_bits,
                               }
        platform_quant = platform_to_string(
            config.PlatFormQuant.platform_quant)
        op_inner = None
        input_list = None
        input_back_list = None
        if is_not_from_iqtensor:
            op_inner = quantlinear(g, input, scale_i(),
                                   platform_quant, data_bits)
            input_list = [op_inner, weight_ih, weight_hh]
            input_back_list = [op_inner, weight_ih_reverse, weight_hh_reverse]
        else:
            input_list = [input, weight_ih, weight_hh]
            input_back_list = [input, weight_ih_reverse, weight_hh_reverse]
        if bias_ih is not None and bias_hh is not None:
            input_list.append(bias_ih)
            input_list.append(bias_hh)
            input_back_list.append(bias_ih_reverse)
            input_back_list.append(bias_hh_reverse)
        if o_bits is not None:
            param_dict['scale_o_f'] = scale_o()
            param_dict['o_bits_i'] = o_bits
            if bidirectional:
                param_back_dict['scale_o_f'] = scale_o_reverse()
                param_back_dict['o_bits_i'] = o_bits

        param_dict['platform_quant_s'] = platform_quant
        param_dict['outputs'] = 2
        param_back_dict['platform_quant_s'] = platform_quant
        param_back_dict['outputs'] = 2
        if lengths is None and hidden_state is None:
            gru, hidden = g.op("thinker::GRUInt", *input_list, **param_dict)
            if bidirectional:
                gru_backward, hidden_back = g.op(
                    "thinker::GRUInt", *input_back_list, **param_back_dict)
                if o_bits is not None:
                    scale_o = ScalerBuffer(min(scale_o(), scale_o_reverse()))
                    args = [gru, gru_backward, scale_o, scale_o_reverse]
                    gru = iqcat_sym(g, None, scale_o, None, 2,
                                    None, False, None, None, None, *args)
                else:
                    args = [gru, gru_backward]
                    gru = g.op("Concat", *args, axis_i=2)
                hidden = g.op("Concat", hidden, hidden_back, axis_i=0)
        elif lengths is not None and hidden_state is None:
            input_list.insert(1, lengths)
            input_back_list.insert(1, lengths)
            gru, hidden = g.op("thinker::GRUInt_Is8_Is64",
                               *input_list, **param_dict)
            if bidirectional:
                gru_backward, hidden_back = g.op(
                    "thinker::GRUInt_Is8_Is64", *input_back_list, **param_back_dict)
                if o_bits is not None:
                    scale_o = ScalerBuffer(min(scale_o(), scale_o_reverse()))
                    args = [gru, gru_backward, scale_o, scale_o_reverse]
                    gru = iqcat_sym(g, None, scale_o, None, 2,
                                    None, False, None, None, None, *args)
                else:
                    args = [gru, gru_backward]
                    gru = g.op("Concat", *args, axis_i=2)
                hidden = g.op("Concat", hidden, hidden_back, axis_i=0)
        else:
            input_list.insert(1, lengths)
            input_list.insert(2, hidden_state)
            input_back_list.insert(1, lengths)
            input_back_list.insert(2, hidden_state)
            gru, hidden = g.op(
                "thinker::GRUInt_Is8_Is64_If32", *input_list, **param_dict)
            if bidirectional:
                gru_backward, hidden_back = g.op(
                    "thinker::GRUInt_Is8_Is64_If32", *input_back_list, **param_back_dict)
                if o_bits is not None:
                    scale_o = ScalerBuffer(min(scale_o(), scale_o_reverse()))
                    args = [gru, gru_backward, scale_o, scale_o_reverse]
                    gru = iqcat_sym(g, None, scale_o, None, 2,
                                    None, False, None, None, None, *args)
                else:
                    args = [gru, gru_backward]
                    gru = g.op("Concat", *args, axis_i=2)
                hidden = g.op("Concat", hidden, hidden_back, axis_i=0)

        return gru, hidden


class GRUInt(nn.GRU):
    r"""实现GRUInt的量化训练与测试，继承自nn.GRU,

    Args:
        input_size hidden_size num_layers bias batch_first dropout bidirectional
        与nn.GRU一致的参数
        unified(bool): 确认正反向参数统计是否一致
        data_bits(int): 输入量化位数
        parameter_bits(int): 参数量化位数
        mode(Enum): 量化方式，支持MaxValue与Qvalue
        o_bits(int, default=None):输出量化位数
        scale_i(np.float32): 统计的是输入scale，输入大小为(b, t, d)或(t, b, d)
        scale_h(np.float32): 统计的是每一帧计算隐藏输出的最值momentum统计得到的scale
        scale_iw(np.float32): 依据最终的模型参数计算得到，无统计
        scale_hw(np.float32): 依据最终模型参数计算得到，无统计参数
        scale_o(np.float32): 最终输出的统计scale
        scale_reverse_*(np.float32):对应反向过程中各个scale数值
    """

    def __init__(self, input_size, hidden_size, num_layers=1, bias=True, batch_first=False, dropout=0, bidirectional=False, unified=True,
                 data_bits=8, parameter_bits=8, mode=QuantMode.QValue, o_bits=None, clamp_data=None, clamp_weight=None, clamp_bias=None):
        nn.GRU.__init__(self, input_size, hidden_size, num_layers,
                        bias, batch_first, dropout, bidirectional)
        ModuleIntConfig.__init__(
            self, data_bits=data_bits, parameter_bits=parameter_bits, mode=mode, o_bits=o_bits)
        self.prefix = ""
        self.dump = False
        self.path = ""
        self.unified = unified
        self.momentum = 0.1
        self.is_not_from_iqtensor = True
        self.clamp_data = clamp_data
        self.clamp_weight = clamp_weight
        self.clamp_bias = clamp_bias

        self.register_buffer('running_i', torch.zeros(1))
        self.register_buffer('running_h', torch.zeros(1))
        self.register_buffer('running_iw', torch.zeros(1))
        self.register_buffer('running_hw', torch.zeros(1))
        self.register_buffer('running_io', torch.zeros(1))
        self.register_buffer('running_ho', torch.zeros(1))
        self.register_buffer('running_o', torch.zeros(1))

        self.register_buffer('scale_i', torch.zeros(1))
        self.register_buffer('scale_h', torch.zeros(1))
        self.register_buffer('scale_iw', torch.zeros(1))
        self.register_buffer('scale_hw', torch.zeros(1))
        self.register_buffer('scale_io', torch.zeros(1))
        self.register_buffer('scale_ho', torch.zeros(1))
        self.register_buffer('scale_o', torch.zeros(1))
        if self.bidirectional:
            self.register_buffer('running_i_reverse', torch.zeros(1))
            self.register_buffer('running_h_reverse', torch.zeros(1))
            self.register_buffer('running_iw_reverse', torch.zeros(1))
            self.register_buffer('running_hw_reverse', torch.zeros(1))
            self.register_buffer('running_io_reverse', torch.zeros(1))
            self.register_buffer('running_ho_reverse', torch.zeros(1))
            self.register_buffer('running_o_reverse', torch.zeros(1))

            self.register_buffer('scale_i_reverse', torch.zeros(1))
            self.register_buffer('scale_h_reverse', torch.zeros(1))
            self.register_buffer('scale_iw_reverse', torch.zeros(1))
            self.register_buffer('scale_hw_reverse', torch.zeros(1))
            self.register_buffer('scale_io_reverse', torch.zeros(1))
            self.register_buffer('scale_ho_reverse', torch.zeros(1))
            self.register_buffer('scale_o_reverse', torch.zeros(1))
        self.sigmoid_table = None
        self.tanh_table = None

    def _single_direction_tensor(self, input, hidden, layer=0, direct=0):
        step_outputs = []
        weight_ih = self.weight_ih_l0_reverse if direct == 1 else self.weight_ih_l0
        weight_hh = self.weight_hh_l0_reverse if direct == 1 else self.weight_hh_l0
        bias_ih = self.bias_ih_l0_reverse if direct == 1 else self.bias_ih_l0
        bias_hh = self.bias_hh_l0_reverse if direct == 1 else self.bias_hh_l0
        if weight_ih.dtype == torch.float32:
            weight_ih = normalize_weight_with_config(
                weight_ih, self.clamp_weight, self.training)
            weight_hh = normalize_weight_with_config(
                weight_hh, self.clamp_weight, self.training)
            bias_ih = normalize_bias_with_config(
                bias_ih, self.clamp_bias, self.training)
            bias_hh = normalize_bias_with_config(
                bias_hh, self.clamp_bias, self.training)

        running_i_tensor = self.running_i_reverse if direct == 1 and not self.unified else self.running_i
        running_h_tensor = self.running_h_reverse if direct == 1 and not self.unified else self.running_h
        running_iw_tensor = self.running_iw_reverse if direct == 1 and not self.unified else self.running_iw
        running_hw_tensor = self.running_hw_reverse if direct == 1 and not self.unified else self.running_hw
        running_io_tensor = self.running_io_reverse if direct == 1 and not self.unified else self.running_io
        running_ho_tensor = self.running_ho_reverse if direct == 1 and not self.unified else self.running_ho
        running_o_tensor = self.running_o_reverse if direct == 1 and not self.unified else self.running_o

        scale_i_tensor = self.scale_i_reverse if direct == 1 and not self.unified else self.scale_i
        scale_h_tensor = self.scale_h_reverse if direct == 1 and not self.unified else self.scale_h
        scale_iw_tensor = self.scale_iw_reverse if direct == 1 and not self.unified else self.scale_iw
        scale_hw_tensor = self.scale_hw_reverse if direct == 1 and not self.unified else self.scale_hw
        scale_io_tensor = self.scale_io_reverse if direct == 1 and not self.unified else self.scale_io
        scale_ho_tensor = self.scale_ho_reverse if direct == 1 and not self.unified else self.scale_ho
        scale_o_tensor = self.scale_o_reverse if direct == 1 and not self.unified else self.scale_o

        running_i = ScalerBuffer(running_i_tensor)
        running_h = ScalerBuffer(running_h_tensor)
        running_iw = ScalerBuffer(running_iw_tensor)
        running_hw = ScalerBuffer(running_hw_tensor)
        running_io = ScalerBuffer(running_io_tensor)
        running_ho = ScalerBuffer(running_ho_tensor)
        running_o = ScalerBuffer(running_o_tensor)
        scale_i = ScalerBuffer(scale_i_tensor)
        scale_h = ScalerBuffer(scale_h_tensor)
        scale_iw = ScalerBuffer(scale_iw_tensor)
        scale_hw = ScalerBuffer(scale_hw_tensor)
        scale_io = ScalerBuffer(scale_io_tensor)
        scale_ho = ScalerBuffer(scale_ho_tensor)
        scale_o = ScalerBuffer(scale_o_tensor)

        input = torch.cat(input.split(1, 0)[::-1]) if direct == 1 else input
        if self.training:
            if self.is_not_from_iqtensor:
                max_value_ix = get_max_value(input)
                running_i.mul_(
                    1-self.momentum).add_(self.momentum*max_value_ix)
        for input_x in input:
            hidden = GRUCellFunction.apply(input_x, hidden, weight_ih, weight_hh, bias_ih, bias_hh,
                                           self.data_bits, self.parameter_bits, self.o_bits,
                                           running_i, running_h, running_iw, running_hw, running_io, running_ho, running_o,
                                           scale_i, scale_h, scale_iw, scale_hw, scale_io, scale_ho, scale_o,
                                           self.momentum, self.training, self.prefix, self.dump, self.path, self.quant_mode, self.quant, self.is_not_from_iqtensor,
                                           self.clamp_data)
            step_outputs.append(hidden)

        running_i_tensor.fill_(running_i())
        running_h_tensor.fill_(running_h())
        running_iw_tensor.fill_(running_iw())
        running_hw_tensor.fill_(running_hw())
        running_io_tensor.fill_(running_io())
        running_ho_tensor.fill_(running_ho())
        running_o_tensor.fill_(running_o())
        scale_i_tensor.fill_(scale_i())
        scale_h_tensor.fill_(scale_h())
        scale_iw_tensor.fill_(scale_iw())
        scale_hw_tensor.fill_(scale_hw())
        scale_io_tensor.fill_(scale_io())
        scale_ho_tensor.fill_(scale_ho())
        scale_o_tensor.fill_(scale_o())

        step_outputs = step_outputs[::-1] if direct == 1 else step_outputs
        output = torch.stack(step_outputs, 0)
        return output, hidden

    def _single_direction_packed(self, input, hidden, layer=0, direct=0, batch_sizes=None):
        if direct:
            return self._packed_reverse(input, hidden, layer, direct, batch_sizes)
        else:
            return self._packed_forward(input, hidden, layer, direct, batch_sizes)

    def _packed_forward(self, input, hidden, layer=0, direct=0, batch_sizes=None):
        step_outputs = []
        final_hiddens = []
        weight_ih = self.weight_ih_l0_reverse if direct == 1 else self.weight_ih_l0
        weight_hh = self.weight_hh_l0_reverse if direct == 1 else self.weight_hh_l0
        bias_ih = self.bias_ih_l0_reverse if direct == 1 else self.bias_ih_l0
        bias_hh = self.bias_hh_l0_reverse if direct == 1 else self.bias_hh_l0
        if weight_ih.dtype == torch.float32:
            weight_ih = normalize_weight_with_config(
                weight_ih, self.clamp_weight, self.training)
            weight_hh = normalize_weight_with_config(
                weight_hh, self.clamp_weight, self.training)
            bias_ih = normalize_bias_with_config(
                bias_ih, self.clamp_bias, self.training)
            bias_hh = normalize_bias_with_config(
                bias_hh, self.clamp_bias, self.training)

        running_i_tensor = self.running_i_reverse if direct == 1 and not self.unified else self.running_i
        running_h_tensor = self.running_h_reverse if direct == 1 and not self.unified else self.running_h
        running_iw_tensor = self.running_iw_reverse if direct == 1 and not self.unified else self.running_iw
        running_hw_tensor = self.running_hw_reverse if direct == 1 and not self.unified else self.running_hw
        running_io_tensor = self.running_io_reverse if direct == 1 and not self.unified else self.running_io
        running_ho_tensor = self.running_ho_reverse if direct == 1 and not self.unified else self.running_ho
        running_o_tensor = self.running_o_reverse if direct == 1 and not self.unified else self.running_o

        scale_i_tensor = self.scale_i_reverse if direct == 1 and not self.unified else self.scale_i
        scale_h_tensor = self.scale_h_reverse if direct == 1 and not self.unified else self.scale_h
        scale_iw_tensor = self.scale_iw_reverse if direct == 1 and not self.unified else self.scale_iw
        scale_hw_tensor = self.scale_hw_reverse if direct == 1 and not self.unified else self.scale_hw
        scale_io_tensor = self.scale_io_reverse if direct == 1 and not self.unified else self.scale_io
        scale_ho_tensor = self.scale_ho_reverse if direct == 1 and not self.unified else self.scale_ho
        scale_o_tensor = self.scale_o_reverse if direct == 1 and not self.unified else self.scale_o

        running_i = ScalerBuffer(running_i_tensor)
        running_h = ScalerBuffer(running_h_tensor)
        running_iw = ScalerBuffer(running_iw_tensor)
        running_hw = ScalerBuffer(running_hw_tensor)
        running_io = ScalerBuffer(running_io_tensor)
        running_ho = ScalerBuffer(running_ho_tensor)
        running_o = ScalerBuffer(running_o_tensor)
        scale_i = ScalerBuffer(scale_i_tensor)
        scale_h = ScalerBuffer(scale_h_tensor)
        scale_iw = ScalerBuffer(scale_iw_tensor)
        scale_hw = ScalerBuffer(scale_hw_tensor)
        scale_io = ScalerBuffer(scale_io_tensor)
        scale_ho = ScalerBuffer(scale_ho_tensor)
        scale_o = ScalerBuffer(scale_o_tensor)

        hidden = copy.deepcopy(hidden)
        input, batch_size_list = _unbind_packed(input, batch_sizes)
        last_batch_size = batch_size_list[0]
        if self.training:
            if self.is_not_from_iqtensor:
                max_value_ix = get_max_value(input)
                running_i.mul_(
                    1-self.momentum).add_(self.momentum*max_value_ix)
        for input_i, batch_len in zip(input, batch_size_list):
            inc = batch_len - last_batch_size
            if inc < 0:
                # record unused-hidden of last-time
                final_hiddens.append(
                    _slice(hidden, batch_len, last_batch_size))
                # slice new hidden
                hidden = hx_slice(None, hidden, last_batch_size, batch_len)
            hidden = GRUCellFunction.apply(input_i, hidden, weight_ih, weight_hh, bias_ih, bias_hh,
                                           self.data_bits, self.parameter_bits, self.o_bits,
                                           running_i, running_h, running_iw, running_hw, running_io, running_ho, running_o,
                                           scale_i, scale_h, scale_iw, scale_hw, scale_io, scale_ho, scale_o,
                                           self.momentum, self.training, self.prefix, self.dump, self.path, self.quant_mode, self.quant,
                                           self.is_not_from_iqtensor, self.clamp_data)
            step_outputs.append(hidden)
            last_batch_size = batch_len

        running_i_tensor.fill_(running_i())
        running_h_tensor.fill_(running_h())
        running_iw_tensor.fill_(running_iw())
        running_hw_tensor.fill_(running_hw())
        running_io_tensor.fill_(running_io())
        running_ho_tensor.fill_(running_ho())
        running_o_tensor.fill_(running_o())
        scale_i_tensor.fill_(scale_i())
        scale_h_tensor.fill_(scale_h())
        scale_iw_tensor.fill_(scale_iw())
        scale_hw_tensor.fill_(scale_hw())
        scale_io_tensor.fill_(scale_io())
        scale_ho_tensor.fill_(scale_ho())
        scale_o_tensor.fill_(scale_o())

        final_hiddens.append(hidden)
        ret_hidden = final_hiddens[::-1]
        hy_list = []
        for each in ret_hidden:
            hy_list.append(each)
        hidden = torch.cat(hy_list, 0)
        output = torch.cat(step_outputs, 0)
        return output, hidden

    def _packed_reverse(self, input, hidden, layer=0, direct=0, batch_sizes=None, bQuant=False):
        step_outputs = []
        weight_ih = self.weight_ih_l0_reverse if direct == 1 else self.weight_ih_l0
        weight_hh = self.weight_hh_l0_reverse if direct == 1 else self.weight_hh_l0
        bias_ih = self.bias_ih_l0_reverse if direct == 1 else self.bias_ih_l0
        bias_hh = self.bias_hh_l0_reverse if direct == 1 else self.bias_hh_l0
        if weight_ih.dtype == torch.float32:
            weight_ih = normalize_weight_with_config(
                weight_ih, self.clamp_weight, self.training)
            weight_hh = normalize_weight_with_config(
                weight_hh, self.clamp_weight, self.training)
            bias_ih = normalize_bias_with_config(
                bias_ih, self.clamp_bias, self.training)
            bias_hh = normalize_bias_with_config(
                bias_hh, self.clamp_bias, self.training)

        running_i_tensor = self.running_i_reverse if direct == 1 and not self.unified else self.running_i
        running_h_tensor = self.running_h_reverse if direct == 1 and not self.unified else self.running_h
        running_iw_tensor = self.running_iw_reverse if direct == 1 and not self.unified else self.running_iw
        running_hw_tensor = self.running_hw_reverse if direct == 1 and not self.unified else self.running_hw
        running_io_tensor = self.running_io_reverse if direct == 1 and not self.unified else self.running_io
        running_ho_tensor = self.running_ho_reverse if direct == 1 and not self.unified else self.running_ho
        running_o_tensor = self.running_o_reverse if direct == 1 and not self.unified else self.running_o

        scale_i_tensor = self.scale_i_reverse if direct == 1 and not self.unified else self.scale_i
        scale_h_tensor = self.scale_h_reverse if direct == 1 and not self.unified else self.scale_h
        scale_iw_tensor = self.scale_iw_reverse if direct == 1 and not self.unified else self.scale_iw
        scale_hw_tensor = self.scale_hw_reverse if direct == 1 and not self.unified else self.scale_hw
        scale_io_tensor = self.scale_io_reverse if direct == 1 and not self.unified else self.scale_io
        scale_ho_tensor = self.scale_ho_reverse if direct == 1 and not self.unified else self.scale_ho
        scale_o_tensor = self.scale_o_reverse if direct == 1 and not self.unified else self.scale_o

        running_i = ScalerBuffer(running_i_tensor)
        running_h = ScalerBuffer(running_h_tensor)
        running_iw = ScalerBuffer(running_iw_tensor)
        running_hw = ScalerBuffer(running_hw_tensor)
        running_io = ScalerBuffer(running_io_tensor)
        running_ho = ScalerBuffer(running_ho_tensor)
        running_o = ScalerBuffer(running_o_tensor)
        scale_i = ScalerBuffer(scale_i_tensor)
        scale_h = ScalerBuffer(scale_h_tensor)
        scale_iw = ScalerBuffer(scale_iw_tensor)
        scale_hw = ScalerBuffer(scale_hw_tensor)
        scale_io = ScalerBuffer(scale_io_tensor)
        scale_ho = ScalerBuffer(scale_ho_tensor)
        scale_o = ScalerBuffer(scale_o_tensor)

        input, batch_size_list = _unbind_packed(input, batch_sizes)
        input = input[::-1]
        batch_size_list = batch_size_list[::-1]
        input_hx = copy.deepcopy(hidden)
        last_batch_size = batch_size_list[0]
        hidden = _slice(hidden, 0, last_batch_size)
        if self.training:
            if self.is_not_from_iqtensor:
                max_value_ix = get_max_value(input)
                running_i.mul_(
                    1-self.momentum).add_(self.momentum*max_value_ix)
        for input_i, batch_len in zip(input, batch_size_list):
            if last_batch_size != batch_len:
                hidden = hx_slice(input_hx, hidden, last_batch_size, batch_len)
            hidden = GRUCellFunction.apply(input_i, hidden, weight_ih, weight_hh, bias_ih, bias_hh,
                                           self.data_bits, self.parameter_bits, self.o_bits,
                                           running_i, running_h, running_iw, running_hw, running_io, running_ho, running_o,
                                           scale_i, scale_h, scale_iw, scale_hw, scale_io, scale_ho, scale_o,
                                           self.momentum, self.training, self.prefix, self.dump, self.path, self.quant_mode, self.quant,
                                           self.is_not_from_iqtensor, self.clamp_data)
            step_outputs.append(hidden)
            last_batch_size = batch_len

        running_i_tensor.fill_(running_i())
        running_h_tensor.fill_(running_h())
        running_iw_tensor.fill_(running_iw())
        running_hw_tensor.fill_(running_hw())
        running_io_tensor.fill_(running_io())
        running_ho_tensor.fill_(running_ho())
        running_o_tensor.fill_(running_o())
        scale_i_tensor.fill_(scale_i())
        scale_h_tensor.fill_(scale_h())
        scale_iw_tensor.fill_(scale_iw())
        scale_hw_tensor.fill_(scale_hw())
        scale_io_tensor.fill_(scale_io())
        scale_ho_tensor.fill_(scale_ho())
        scale_o_tensor.fill_(scale_o())

        step_outputs = step_outputs[::-1]
        output = torch.cat(step_outputs, 0)
        return output, hidden

    def _finetune(self, input, hidden, layer=0, direct=0, batch_sizes=None):
        if batch_sizes is None:
            return self._single_direction_tensor(input, hidden, layer, direct)
        else:
            return self._single_direction_packed(input, hidden, layer, direct, batch_sizes)

    def _run_single_direction(self, input, hidden, layer=0, direct=0, batch_sizes=None):
        return self._finetune(input, hidden, layer, direct, batch_sizes)

    def single_direction(self, input, layer, hx, batch_sizes=None):
        hidden = hx
        output, hidden = self._run_single_direction(
            input, hidden, layer, direct=0, batch_sizes=batch_sizes)
        return output, [hidden]

    def bidirection(self, input, layer, hx, batch_sizes=None):
        hx_f = hx[0]
        hx_b = hx[1]
        fw_output, fw_hidden = self._run_single_direction(
            input, hx_f, layer, direct=0, batch_sizes=batch_sizes)
        rev_output, rev_hidden = self._run_single_direction(
            input, hx_b, layer, direct=1, batch_sizes=batch_sizes)
        if batch_sizes is None:
            output = torch.cat((fw_output, rev_output), fw_output.dim()-1)
        else:  # packed sequence
            output = torch.cat((fw_output, rev_output), -1)
        return output, [fw_hidden, rev_hidden]

    def gru_forward(self, input, hiddens, batch_sizes=None):
        final_hiddens = []
        for layer_num in range(self.num_layers):
            hid = hiddens[layer_num] if hiddens is not None else None
            output, hc = self.bidirection(input, layer_num, hid, batch_sizes) if self.bidirectional else self.single_direction(
                input, layer_num, hid, batch_sizes)
            final_hiddens.extend(hc)
            input = output
            # add dropout
            if (self.dropout != 0 and self.training and layer_num < self.num_layers - 1):
                input = torch.nn.functional.dropout(input, self.dropout)

        hy = [hidden for hidden in final_hiddens]
        hy = torch.stack(hy, 0)

        return input, hy

    def _generate_hiddens(self, hx):
        if hx is not None:
            hidden_list = _unbind(hx)
            length = len(hidden_list)
            if self.bidirectional:
                assert length/self.num_layers % 2 == 0, 'hidden len must be double in bidirectional mode'
            i = 0
            hiddens = []
            while i < length:
                if self.bidirectional:
                    hiddens.append((hidden_list[i], hidden_list[i+1]))
                    i += 2
                else:
                    hiddens.append(hidden_list[i])
                    i += 1
        else:
            hiddens = None
        return hiddens

    def forward_input_tensor(self, input, hx, batch_sizes=None):
        input = input.transpose(0, 1) if self.batch_first else input
        hiddens = self._generate_hiddens(hx)
        output, hr = self.gru_forward(input, hiddens)
        output = output.transpose(0, 1) if self.batch_first else output
        return output, hr

    def forward_input_packed(self, input, hx, batch_sizes=None):
        hiddens = self._generate_hiddens(hx)
        output, hr = self.gru_forward(input, hiddens, batch_sizes)
        return output, hr

    def forward(self, input, hx=None):
        orig_input = input
        if not is_in_onnx_export():
            if isinstance(orig_input, tuple):
                input, lengths, batch_first, enforce_sorted = orig_input
                if isinstance(input, IQTensor):
                    self.is_not_from_iqtensor = False
                    if input.bits != self.data_bits:
                        input = Requant.apply(
                            input, input.bits, input.scale_data, self.data_bits)
                    self.scale_i.fill_(input.scale_data)
                    self.running_i.fill_(input.running_data)
                    if self.bidirectional:
                        self.scale_i_reverse.fill_(input.scale_data)
                        self.running_i_reverse.fill_(input.running_data)
                packed_input = torch_pack_padded_sequence(
                    input, lengths, batch_first, enforce_sorted)
                input, batch_sizes, sorted_indices, unsorted_indices = packed_input
                max_batch_size = batch_sizes[0]
                max_batch_size = int(max_batch_size)
            else:
                batch_sizes = None
                max_batch_size = input.size(
                    0) if self.batch_first else input.size(1)
                sorted_indices = None
                unsorted_indices = None
                if isinstance(input, IQTensor):
                    self.is_not_from_iqtensor = False
                    if input.bits != self.data_bits:
                        input = Requant.apply(
                            input, input.bits, input.scale_data, self.data_bits)
                    self.scale_i.fill_(input.scale_data)
                    self.running_i.fill_(input.running_data)
                    if self.bidirectional:
                        self.scale_i_reverse.fill_(input.scale_data)
                        self.running_i_reverse.fill_(input.running_data)
            assert self.num_layers == 1, 'invalid num_layers, now only support num_layers = 1'
            if hx is None:
                num_directions = 2 if self.bidirectional else 1
                hx = torch.zeros(self.num_layers * num_directions,
                                 max_batch_size, self.hidden_size,
                                 dtype=input.dtype, device=input.device)
            else:
                # Each batch of the hidden state should match the input sequence that
                # the user believes he/she is passing in.
                hx = self.permute_hidden(hx, sorted_indices)

            if batch_sizes is not None:
                output, hidden = self.forward_input_packed(
                    input, hx, batch_sizes)
            else:
                output, hidden = self.forward_input_tensor(input, hx)
            if isinstance(orig_input, tuple):
                output_packed = PackedSequence(
                    output, batch_sizes, sorted_indices, unsorted_indices)
                output, lengths = torch_pad_packed_sequence(
                    output_packed, self.batch_first)
                if self.o_bits is not None:
                    if self.training:
                        output = Quant2IQTensor.apply(
                            output, self.o_bits, self.quant_mode, 'output')
                    else:
                        scale_o = ScalerBuffer(self.quant.running_to_scale(
                            self.running_o, self.o_bits, mode=self.quant_mode))
                        if self.bidirectional:
                            if self.unified:
                                scale_o_reverse = scale_o
                            else:
                                scale_o_reverse = ScalerBuffer(self.quant.running_to_scale(
                                    self.running_o_reverse, self.o_bits, mode=self.quant_mode))
                            scale_o = ScalerBuffer(
                                min(scale_o(), scale_o_reverse()))
                        output = from_torch_tensor(
                            output, scale_o(), self.o_bits)
                return (output, lengths), self.permute_hidden(hidden, unsorted_indices)
            else:
                if self.o_bits is not None:
                    if self.training:
                        output = Quant2IQTensor.apply(
                            output, self.o_bits, self.quant_mode, 'output')
                    else:
                        scale_o = ScalerBuffer(self.quant.running_to_scale(
                            self.running_o, self.o_bits, mode=self.quant_mode))
                        if self.bidirectional:
                            if self.unified:
                                scale_o_reverse = scale_o
                            else:
                                scale_o_reverse = ScalerBuffer(self.quant.running_to_scale(
                                    self.running_o_reverse, self.o_bits, mode=self.quant_mode))
                            scale_o = ScalerBuffer(
                                min(scale_o(), scale_o_reverse()))
                        output = from_torch_tensor(
                            output, scale_o(), self.o_bits)
                return output, self.permute_hidden(hidden, unsorted_indices)
        else:
            lengths = None
            if isinstance(orig_input, tuple):
                input, lengths, _, _ = orig_input
            else:
                input = orig_input
                lengths = None
            if isinstance(input, IQTensor):
                self.is_not_from_iqtensor = False
                if input.bits != self.data_bits:
                    input = Requant.apply(
                        input, input.bits, input.scale_data, self.data_bits)
            bias_ih = None
            bias_hh = None
            bias_ih_reverse = None
            bias_hh_reverse = None
            weight_ih = self.weight_ih_l0
            weight_hh = self.weight_hh_l0
            weight_ih_reverse = weight_ih
            weight_hh_reverse = weight_hh
            if self.bias:
                bias_ih = self.bias_ih_l0
                bias_hh = self.bias_hh_l0
                bias_ih_reverse = bias_ih
                bias_hh_reverse = bias_hh
            if self.bidirectional:
                weight_ih_reverse = self.weight_ih_l0_reverse
                weight_hh_reverse = self.weight_hh_l0_reverse
                bias_ih_reverse = self.bias_ih_l0_reverse
                bias_hh_reverse = self.bias_hh_l0_reverse

            scale_i = ScalerBuffer(self.scale_i)
            scale_iw = ScalerBuffer(self.scale_iw)
            scale_io = ScalerBuffer(self.scale_io)
            scale_h = ScalerBuffer(self.scale_h)
            scale_hw = ScalerBuffer(self.scale_hw)
            scale_ho = ScalerBuffer(self.scale_ho)
            scale_o = ScalerBuffer(self.scale_o)
            scale_i_reverse = None
            scale_iw_reverse = None
            scale_io_reverse = None
            scale_h_reverse = None
            scale_hw_reverse = None
            scale_ho_reverse = None
            scale_o_reverse = None
            hidden_state = None
            cell_state = None
            if self.bidirectional:
                scale_i_reverse = ScalerBuffer(self.scale_i_reverse)
                scale_iw_reverse = ScalerBuffer(self.scale_iw_reverse)
                scale_io_reverse = ScalerBuffer(self.scale_io_reverse)
                scale_h_reverse = ScalerBuffer(self.scale_h_reverse)
                scale_hw_reverse = ScalerBuffer(self.scale_hw_reverse)
                scale_ho_reverse = ScalerBuffer(self.scale_ho_reverse)
                scale_o_reverse = ScalerBuffer(self.scale_o_reverse)
            output = None
            hy = None
            if hx is not None:
                hidden_state = hx
            if hidden_state is not None:
                batch_size = input.size(
                    0) if self.batch_first else input.size(1)
                seq_len = input.size(
                    1) if self.batch_first else input.size(0)
                lengths = torch.tensor([seq_len for i in range(
                    batch_size)], dtype=torch.int64, device=input.device) if lengths is None else lengths
            output, hy = GRUSingleONNXFunction.apply(input, lengths, hidden_state, weight_ih, weight_hh, bias_ih, bias_hh,
                                                     weight_ih_reverse, weight_hh_reverse, bias_ih_reverse, bias_hh_reverse,
                                                     self.input_size, self.hidden_size, self.num_layers, self.batch_first, self.dropout, self.bidirectional,
                                                     self.data_bits, self.parameter_bits, self.o_bits,
                                                     scale_i, scale_iw, scale_io, scale_h, scale_hw, scale_ho, scale_o,
                                                     scale_i_reverse, scale_iw_reverse, scale_io_reverse, scale_h_reverse, scale_hw_reverse, scale_ho_reverse, scale_o_reverse,
                                                     self.is_not_from_iqtensor)
            if self.o_bits is not None:
                if self.bidirectional:
                    scale_o = ScalerBuffer(min(scale_o(), scale_o_reverse()))
                output = from_torch_tensor(output, scale_o(), self.o_bits)
            if isinstance(orig_input, tuple):
                return (output, lengths), hy
            else:
                return output, hy

    def _load_from_state_dict(self, state_dict, prefix, local_metadata, strict,
                              missing_keys, unexpected_keys, error_msgs):
        ModuleIntConfig._load_from_state_dict_global(
            self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs)

    def state_dict(self, destination=None, prefix='', keep_vars=False):
        if destination is None:
            destination = OrderedDict()
            destination._metadata = OrderedDict()
        destination._metadata[prefix[:-1]
                              ] = local_metadata = dict(version=self._version)
        if is_in_onnx_export():
            assert self.running_i > 0, 'invalid running_x <= 0'
            scale_i = ScalerBuffer(self.scale_i.data)
            if self.is_not_from_iqtensor:
                scale_i = ScalerBuffer(self.quant.running_to_scale(
                    self.running_i, self.data_bits, mode=self.quant_mode))
                self.scale_i.data.fill_(scale_i())
            scale_h = ScalerBuffer(self.quant.running_to_scale(
                self.running_h, self.data_bits, mode=self.quant_mode))
            self.scale_h.data.fill_(scale_h())
            if self.o_bits is not None:
                scale_o = ScalerBuffer(self.quant.running_to_scale(
                    self.running_o, self.o_bits, mode=self.quant_mode))
                self.scale_o.data.fill_(scale_o())

            if self.bidirectional:
                if self.unified:
                    self.running_i_reverse.data = self.running_i.data
                    self.running_h_reverse.data = self.running_h.data
                    self.scale_i_reverse.data = self.scale_i.data
                    self.scale_h_reverse.data = self.scale_h.data
                    scale_i_reverse = scale_i
                    scale_h_reverse = scale_h
                    if self.o_bits is not None:
                        self.running_o_reverse.data = self.running_o.data
                        self.scale_o_reverse.data = self.scale_o.data
                        scale_o_reverse = scale_o
                else:
                    scale_i_reverse = ScalerBuffer(self.scale_i_reverse.data)
                    if self.is_not_from_iqtensor:
                        scale_i_reverse = ScalerBuffer(self.quant.running_to_scale(
                            self.running_i_reverse, self.data_bits, mode=self.quant_mode))
                        self.scale_i_reverse.data.fill_(scale_i_reverse())
                    scale_h_reverse = ScalerBuffer(self.quant.running_to_scale(
                        self.running_h_reverse, self.data_bits, mode=self.quant_mode))
                    self.scale_h_reverse.data.fill_(scale_h_reverse())
                    if self.o_bits is not None:
                        scale_o_reverse = ScalerBuffer(self.quant.running_to_scale(
                            self.running_o_reverse, self.o_bits, mode=self.quant_mode))
                        self.scale_o_reverse.data.fill_(scale_o_reverse())

        if self.weight_ih_l0.dtype == torch.float32:
            clamp_weight_iw = normalize_weight_with_config(
                self.weight_ih_l0, self.clamp_weight, False)
            clamp_weight_hw = normalize_weight_with_config(
                self.weight_hh_l0, self.clamp_weight, False)
            self.weight_ih_l0.data = clamp_weight_iw
            self.weight_hh_l0.data = clamp_weight_hw
            if is_in_onnx_export():
                scale_iw = ScalerBuffer(self.quant.running_to_scale(
                    self.running_iw, self.parameter_bits, mode=self.quant_mode))
                self.scale_iw.data.fill_(scale_iw())
                scale_hw = ScalerBuffer(self.quant.running_to_scale(
                    self.running_hw, self.parameter_bits, mode=self.quant_mode))
                self.scale_hw.data.fill_(scale_hw())
                q_weight_iw, scale_iw, _ = self.quant.quant(
                    clamp_weight_iw, self.parameter_bits, scale=scale_iw, mode=self.quant_mode, quant_data='weight')
                q_weight_hw, scale_hw, _ = self.quant.quant(
                    clamp_weight_hw, self.parameter_bits, scale=scale_hw, mode=self.quant_mode, quant_data='weight')
            if self.bias:
                clamp_bias_iw = normalize_bias_with_config(
                    self.bias_ih_l0, self.clamp_bias, False)
                clamp_bias_hw = normalize_bias_with_config(
                    self.bias_hh_l0, self.clamp_bias, False)
                self.bias_ih_l0.data = clamp_bias_iw
                self.bias_hh_l0.data = clamp_bias_hw
                if is_in_onnx_export():
                    if config.PlatFormQuant.platform_quant == PlatFormQuant.luna_quant:
                        q_bias_iw = (clamp_bias_iw * scale_i *
                                     scale_iw + 0.5).floor()
                        q_bias_hw = (clamp_bias_hw * scale_h *
                                     scale_hw + 0.5).floor()
                        if self.data_bits + self.parameter_bits <= 16:
                            q_bias_iw = q_bias_iw.float().int()
                            q_bias_hw = q_bias_hw.float().int()
                    else:
                        assert False, "linger only support luna quant."
            if self.bidirectional:
                clamp_weight_iw_reverse = normalize_weight_with_config(
                    self.weight_ih_l0_reverse, self.clamp_weight, False)
                clamp_weight_hw_reverse = normalize_weight_with_config(
                    self.weight_hh_l0_reverse, self.clamp_weight, False)
                self.weight_ih_l0_reverse.data = clamp_weight_iw_reverse
                self.weight_hh_l0_reverse.data = clamp_weight_hw_reverse
                if is_in_onnx_export():
                    if self.unified:
                        q_weight_iw_reverse, scale_iw_reverse, _ = self.quant.quant(
                            clamp_weight_iw_reverse, self.parameter_bits, scale=scale_iw, mode=self.quant_mode, quant_data='weight')
                        q_weight_hw_reverse, scale_hw_reverse, _ = self.quant.quant(
                            clamp_weight_hw_reverse, self.parameter_bits, scale=scale_hw, mode=self.quant_mode, quant_data='weight')
                    else:
                        scale_iw_reverse = ScalerBuffer(self.quant.running_to_scale(
                            self.running_iw_reverse, self.parameter_bits, mode=self.quant_mode))
                        scale_hw_reverse = ScalerBuffer(self.quant.running_to_scale(
                            self.running_hw_reverse, self.parameter_bits, mode=self.quant_mode))
                        q_weight_iw_reverse, scale_iw_reverse, _ = self.quant.quant(
                            clamp_weight_iw_reverse, self.parameter_bits, scale=scale_iw_reverse(), mode=self.quant_mode, quant_data='weight')
                        q_weight_hw_reverse, scale_hw_reverse, _ = self.quant.quant(
                            clamp_weight_hw_reverse, self.parameter_bits, scale=scale_hw_reverse(), mode=self.quant_mode, quant_data='weight')
                    self.scale_iw_reverse.data.fill_(scale_iw_reverse())
                    self.scale_hw_reverse.data.fill_(scale_hw_reverse())
                if self.bias:
                    clamp_bias_iw_reverse = normalize_bias_with_config(
                        self.bias_ih_l0_reverse, self.clamp_bias, False)
                    clamp_bias_hw_reverse = normalize_bias_with_config(
                        self.bias_hh_l0_reverse, self.clamp_bias, False)
                    self.bias_ih_l0_reverse.data = clamp_bias_iw_reverse
                    self.bias_hh_l0_reverse.data = clamp_bias_hw_reverse
                    if is_in_onnx_export():
                        if config.PlatFormQuant.platform_quant == PlatFormQuant.luna_quant:
                            q_bias_iw_reverse = (
                                clamp_bias_iw_reverse * scale_i_reverse * scale_iw_reverse + 0.5).floor()
                            q_bias_hw_reverse = (
                                clamp_bias_hw_reverse * scale_h_reverse * scale_hw_reverse + 0.5).floor()
                            if self.data_bits + self.parameter_bits <= 16:
                                q_bias_iw_reverse = q_bias_iw_reverse.float().int()
                                q_bias_hw_reverse = q_bias_hw_reverse.float().int()
                        else:
                            assert False, "linger only support luna quant."
            if is_in_onnx_export():
                if self.parameter_bits <= 8:
                    self.weight_ih_l0.data = q_weight_iw.char()
                    self.weight_hh_l0.data = q_weight_hw.char()
                    if self.bidirectional:
                        self.weight_ih_l0_reverse.data = q_weight_iw_reverse.char()
                        self.weight_hh_l0_reverse.data = q_weight_hw_reverse.char()
                elif self.parameter_bits <= 16:
                    self.weight_ih_l0.data = q_weight_iw.short()
                    self.weight_hh_l0.data = q_weight_hw.short()
                    if self.bidirectional:
                        self.weight_ih_l0_reverse.data = q_weight_iw_reverse.short()
                        self.weight_hh_l0_reverse.data = q_weight_hw_reverse.short()
                else:
                    self.weight_ih_l0.data = q_weight_iw.int()
                    self.weight_hh_l0.data = q_weight_hw.int()
                    if self.bidirectional:
                        self.weight_ih_l0_reverse.data = q_weight_iw_reverse.int()
                        self.weight_hh_l0_reverse.data = q_weight_hw_reverse.int()
                if self.bias:
                    self.bias_ih_l0.data = q_bias_iw.int()
                    self.bias_hh_l0.data = q_bias_hw.int()
                    if self.bidirectional:
                        self.bias_ih_l0_reverse.data = q_bias_iw_reverse.int()
                        self.bias_hh_l0_reverse.data = q_bias_hw_reverse.int()
        self._save_to_state_dict(destination, prefix, keep_vars)
        for name, self in self._modules.items():
            if self is not None:
                self.state_dict(destination, prefix + name +
                                '.', keep_vars=keep_vars)
        for hook in self._state_dict_hooks.values():
            hook_result = hook(self, destination, prefix, local_metadata)
            if hook_result is not None:
                destination = hook_result
        return destination

    def extra_repr(self):
        s = nn.GRU.extra_repr(self)
        extra_s = ',clamp_data:{clamp_data},clamp_weight:{clamp_weight},clamp_bias:{clamp_bias}'.format(
            **self.__dict__)
        extra_s += ',data_bits:{data_bits},parameter_bits:{parameter_bits},o_bits:{o_bits}'.format(
            **self.__dict__)
        return s+extra_s



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
import linger
import numpy
import torch
import torch.nn as nn


def test_embeddingint_net():
    class Net(nn.Module):
        def __init__(self):
            super(Net, self).__init__()
            self.gather = nn.Embedding(10, 3)

        def forward(self, input):

            x = self.gather(input)
            x = x * 1

            return x
    torch.manual_seed(1)
    torch.cuda.manual_seed_all(1)
    numpy.random.seed(1)
    aa = torch.LongTensor([[1, 2, 4, 5], [4, 3, 2, 9]]).cuda()
    replace_tuple = (nn.Conv2d, nn.LSTM, nn.Embedding)

    net = Net().cuda()
    net = linger.init(net, quant_modules=replace_tuple, mode=linger.QuantMode.QValue)
    criterion = nn.CrossEntropyLoss(ignore_index=-1)
    print(net)
    optimizer = torch.optim.Adam(net.parameters(), lr=0.1)
    loss = None
    label = torch.randint(1, (2,  3)).cuda()
    for i in range(10):
        optimizer.zero_grad()
        out = net(aa)
        loss = criterion(out, label)
        if i % 1 == 0:
            print('loss: ', loss)

        loss.backward()
        optimizer.step()
    with torch.no_grad():
        net.eval()
        net(aa)
        torch.onnx.export(net, (aa), "data.ignore/embedding_int.onnx", export_params=True,
                          opset_version=12, operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)


import math

import torch
import torch.nn as nn
import torch.nn.functional as F

from ..config import config
from ..quant import (Quant, normalize_bias_with_config,
                     normalize_data_with_config, normalize_weight_with_config)
from ..utils import Dump, PlatFormQuant, QuantMode, ScalerBuffer
from .iqtensor import (IQTensor, from_torch_tensor, platform_to_string,
                       quantlinear)
from .ops import ModuleIntConfig
from .requant import Requant


class ConvFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input, weights, bias, kernel_size, padding, stride, dilation, groups, params,
                data_bits, parameter_bits, training, momentum, running_x, running_w, running_o, eval_scale_x, eval_scale_w, eval_scale_o,
                prefix, dump, path, mode, o_bits, quant, ahead_relu, is_not_from_iqtensor, clamp_data, clamp_weight, clamp_bias, ahead_sigmoid):
        scale_o = None

        # venus limits
        # assert input.shape[2] >= weights.shape[2] and input.shape[3] >= weights.shape[
        #     3], f"in Conv2dInt op, input's width >= weight's width && input'height >= weight'height, but you have {input.shape[2]} >= {weights.shape[2]} and {input.shape[3]} >= {weights.shape[3]}"
        assert data_bits in (
            4, 8), f"in Conv2dInt op, data_bits only support 4/8, but you have data_bits {data_bits}"
        assert parameter_bits in (
            4, 8), f"in Conv2dInt op, parameter_bits only support 4/8, but you have parameter_bits {parameter_bits}"
        assert o_bits in (
            4, 8, 16, 32), f"in Conv2dInt op, o_bits only support 4/8/16/32, but you have o_bits {o_bits}"

        channel_in = weights.shape[1]/groups
        # assert not (math.ceil(channel_in/8/stride[1]) * (8*stride[1]) * math.ceil(input.shape[3]/8)*8*1 > 64 * 1024 and channel_in > 512) or not (math.ceil(channel_in/8/stride[1]) * (8*stride[1]) * math.ceil(8/8)*8*input.shape[2] > 64 * 1024 and channel_in >
        #                                                                                                                                           512), f"in Conv2dInt op, the size of the input data after alignment exceed 64KB and channal_in > 512 at the same time is not allowed, but you have (math.ceil({channel_in}/8/{stride[1]}) * (8*{stride[1]}) * math.ceil({input.shape[3]}/8)*8*{1} > 64 * 1024 and {channel_in} > 512) or (math.ceil({channel_in}/8/{stride[1]}) * (8*{stride[1]}) * math.ceil({8}/8)*8*{input.shape[2]} > 64 * 1024 and {channel_in} > 512"

        if training:
            ctx.clamp_data = clamp_data
            ctx.clamp_weight = clamp_weight
            ctx.clamp_bias = clamp_bias
            zero_point = input.zero_point if isinstance(input, IQTensor) else 0
            is_iq_tensor = True if isinstance(input, IQTensor) else False
            ctx.bits = data_bits, parameter_bits, o_bits
            ctx.value = zero_point, is_iq_tensor

            saved_tensors = [input, weights, bias, params]
            if isinstance(input, IQTensor):
                q_input, _, max_value_x = quant.quant(
                    input.data, data_bits, eval_scale_x, mode=QuantMode.QValue, quant_data='input',
                    iq_zero_point=input.zero_point)
                scale_x = eval_scale_x
            else:
                q_input, scale_x, max_value_x = quant.quant(
                    input.data, data_bits, mode=mode, quant_data='input')
                running_x.mul_(1-momentum).add_(momentum*max_value_x)

            q_weights, scale_w, max_value_w = quant.quant(
                weights, parameter_bits, mode=mode, quant_data='weight')
            running_w.mul_(1-momentum).add_(momentum*max_value_w)
            q_input = q_input.float() if data_bits + parameter_bits <= 16 else q_input.double()
            q_weights = q_weights.float() if data_bits + \
                parameter_bits <= 16 else q_weights.double()
            q_outputs = F.conv2d(q_input, q_weights, stride=stride,
                                 padding=padding, dilation=dilation, groups=groups)
            outputs = None
            q_bias = None
            if config.PlatFormQuant.platform_quant in (PlatFormQuant.luna_quant,):
                assert mode == QuantMode.QValue, 'castor quant only support QValue and o_bits=None'
                if bias is not None:
                    q_bias = (bias * scale_w * scale_x + 0.5).floor()
                    if data_bits + parameter_bits <= 16:
                        q_bias = q_bias.float().round()
                    else:
                        q_bias = q_bias.double()
                    # 小心计算溢出问题，未做保护，bias不超过2**23-1数值
                    q_outputs += q_bias.reshape(1, -1, 1, 1)
                outputs = quant.dequant(q_outputs, scale_x*scale_w)
            else:
                assert False, "linger only support luna quant."

            # f_input = quant.dequant(q_input, scale_x)
            # f_weights = quant.dequant(q_weights, scale_w)
            # f_bias = None if bias is None else quant.dequant(q_bias, scale_x*scale_w)
            # saved_tensors = [f_input, f_weights, f_bias, params]

            out_tensor = outputs
            scale_o = None
            if o_bits is not None:
                outputs = normalize_data_with_config(outputs, clamp_data)
                out_tensor = outputs
                q_outputs, scale_o, max_value_o = quant.quant(
                    outputs, o_bits, mode=mode, quant_data='conv_output', ahead_relu=ahead_relu)
                outputs = quant.dequant(q_outputs, scale_o)
                running_o.mul_(1-momentum).add_(momentum*max_value_o)
            ctx.scale = scale_x, scale_w, scale_o
            ctx.save_for_backward(*saved_tensors)

        else:
            assert running_x > 0, 'invalid running_x = 0, please finetune training before eval'
            scale_x = None
            scale_w = None
            scale_o = eval_scale_o
            q_weights = None
            if weights.dtype == torch.float32:
                scale_x = ScalerBuffer(quant.running_to_scale(
                    running_x, data_bits, mode=mode))
                q_weights, scale_w, _ = quant.quant(
                    weights, parameter_bits, mode=mode, quant_data='weight')
                scale_w = ScalerBuffer(scale_w)
                if o_bits is not None:
                    assert running_o > 0, 'invalid running_o <= 0, please finetune training'
                    scale_o = ScalerBuffer(quant.running_to_scale(
                        running_o, o_bits, mode=mode))
            else:
                scale_x = eval_scale_x
                scale_w = eval_scale_w
                q_weights = weights.double()
                if o_bits is not None:
                    scale_o = eval_scale_o

            if isinstance(input, IQTensor):
                scale_x = eval_scale_x
                q_input, _, _ = quant.quant(
                    input.data, data_bits, scale_x, mode=QuantMode.QValue, quant_data='input', iq_zero_point=input.zero_point)
            else:
                q_input, scale_x, _ = quant.quant(
                    input.data, data_bits, scale_x, mode=mode, quant_data='input')
            q_input = q_input.double()
            q_weights = q_weights.double()
            q_outputs = F.conv2d(q_input, q_weights, stride=stride,
                                 padding=padding, dilation=dilation, groups=groups)

            outputs = None
            q_bias = None

            if config.PlatFormQuant.platform_quant in (PlatFormQuant.luna_quant,):
                assert mode == QuantMode.QValue, 'castor quant only support QValue and o_bits=None'
                if bias is not None:
                    if bias.dtype == torch.float32:
                        q_bias = (bias * scale_w * scale_x + 0.5).floor()
                        if data_bits + parameter_bits <= 16:
                            # 降低bias精度，保持训练一致
                            q_bias = q_bias.float().round().double()
                        else:
                            q_bias = q_bias.double()
                    else:
                        q_bias = bias.double()
                    q_outputs += q_bias.reshape(1, -1, 1, 1)
                outputs = quant.dequant(q_outputs, scale_x*scale_w)
            else:
                assert False, "linger only support luna quant."

            if o_bits is not None:
                if config.PlatFormQuant.platform_quant == PlatFormQuant.luna_quant:
                    q_outputs, _, _ = quant.quant(
                        outputs, o_bits, scale_o, mode=mode, quant_data='output')
                else:
                    assert False, "linger only support luna quant."
                outputs = quant.dequant(q_outputs, scale_o)
            if dump:
                if q_bias is not None:
                    name_list = ["input", "weights", "bias", "outputs", "q_input",  "q_weights", "q_bias",
                                 "q_outputs", "scale_x", "scale_w", "scale_o", "running_x", "running_w", "running_o"]
                    attr_list = [input, weights, bias, outputs, q_input, q_weights, q_bias, q_outputs,
                                 scale_x.data, scale_w.data, scale_o.data, running_x.data, running_w.data, running_o.data]
                    Dump.dump_file(prefix, ".Conv2dInt.",
                                   zip(name_list, attr_list), path)
                else:
                    name_list = ["input", "weights", "outputs", "q_input",  "q_weights", "q_outputs",
                                 "scale_x", "scale_w", "scale_o", "running_x", "running_w", "running_o"]
                    attr_list = [input, weights, outputs, q_input, q_weights, q_outputs, scale_x.data,
                                 scale_w.data, scale_o.data, running_x.data, running_w.data, running_o.data]
                    Dump.dump_file(prefix, ".Conv2dInt.",
                                   zip(name_list, attr_list), path)
        if o_bits is None:
            return outputs
        elif isinstance(scale_o, float):
            return from_torch_tensor(outputs, scale_o, o_bits)
        elif isinstance(scale_o, torch.Tensor):
            return from_torch_tensor(outputs, scale_o.item(), o_bits)
        else:
            return from_torch_tensor(outputs, scale_o.data, o_bits)

    @staticmethod
    def backward(ctx, gradOutput):
        clamp_data = ctx.clamp_data
        data_bits, parameter_bits, o_bits = ctx.bits
        zero_point, is_iq_tensor = ctx.value
        scale_x, scale_w, scale_o = ctx.scale
        input, weights, bias, params = ctx.saved_tensors

        if is_iq_tensor:
            f_input = input.data
        else:
            q_input, _, _ = Quant.quant(
                input.data, data_bits, scale_x, mode=QuantMode.QValue, quant_data='input')
            f_input = Quant.dequant(q_input, scale_x)
        f_input = f_input.detach().clone().requires_grad_(True)
        q_weights, _, _ = Quant.quant(
            weights.data, parameter_bits, scale_w, mode=QuantMode.QValue, quant_data='weight')
        f_weights = Quant.dequant(q_weights, scale_w)
        f_weights = f_weights.detach().clone().requires_grad_(True)
        bias = None if bias is None else bias.detach().clone().requires_grad_(True)

        stride = int(params[0]), int(params[1])
        padding = int(params[2]), int(params[3])
        dilation = int(params[4]), int(params[5])
        groups = int(params[6])
        gradBias = None
        # input_clamp = input
        with torch.enable_grad():
            z = F.conv2d(f_input, f_weights, bias,
                         stride, padding, dilation, groups)
            if o_bits is not None:
                z = normalize_data_with_config(z, clamp_data)
            if bias is not None:
                gradInput, gradWeight, gradBias = torch.autograd.grad(
                    z, (f_input, f_weights, bias), gradOutput)
            else:
                gradInput, gradWeight, = torch.autograd.grad(
                    z, (f_input, f_weights), gradOutput)

        return gradInput, gradWeight, gradBias, None, None, None, None, None, None, None, None, None, None, None, None, None,\
            None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None

    @staticmethod
    def symbolic(g, input, weights, bias, kernel_size, padding, stride, dilation, groups, params,
                 data_bits, parameter_bits, training, momentum, running_x, running_w, running_o, scale_x, scale_w, scale_o,
                 prefix, dump, path, mode, o_bits, quant, ahead_relu, is_not_from_iqtensor, clamp_data, clamp_weight, clamp_bias, ahead_sigmoid):
        op_inner = None
        if is_not_from_iqtensor:
            platform_quant = platform_to_string(
                config.PlatFormQuant.platform_quant)
            op_inner = quantlinear(g, input, scale_x(),
                                   platform_quant, data_bits)
        paddings = padding + padding
        param_dict = {'scale_x_f': scale_x(), 'scale_w_f': scale_w(),
                      'dilations_i': dilation, 'group_i': groups, 'kernel_shape_i': kernel_size, 'pads_i': paddings, 'strides_i': stride,
                      'data_bits_i': data_bits, 'parameter_bits_i': parameter_bits}
        if is_not_from_iqtensor:
            input_list = [op_inner, weights]
        else:
            input_list = [input, weights]
        if bias is not None:
            input_list.append(bias)
        platform_quant = platform_to_string(
            config.PlatFormQuant.platform_quant)
        if o_bits is not None:
            param_dict['scale_o_f'] = scale_o()
            param_dict['o_bits_i'] = o_bits
        param_dict['platform_quant_s'] = platform_quant
        if config.PlatFormQuant.platform_quant == PlatFormQuant.luna_quant:
            param_dict['castor_mode_s'] = "luna"
        return g.op("thinker::Conv2dInt", *input_list, **param_dict)


class Conv2dInt(nn.Conv2d, ModuleIntConfig):
    r"""实现Conv2dInt的量化训练与测试，继承自nn.Conv2d,

    Args:
        in_channels out_channels kernel_size stride padding dilation groups bias padding_mode
        与nn.Conv2d一致的参数
        data_bits(int): 输入量化位数
        parameter_bits(int): 参数量化位数
        mode(Enum): 量化方式，支持MaxValue与Qvalue
        o_bits(int, default=None):输出量化位数
        clamp_data(float or None): 针对输出的clamp数值
        clamp_weight(float or None): 针对转为weight的clamp数值
        clamp_bias(float or None): 与clamp_weight一致
        ahead_relu(bool): 是否做融合relu之后的数值统计scale

    """

    def __init__(self, in_channels, out_channels, kernel_size, stride=1,
                 padding=0, dilation=1, groups=1, bias=True, padding_mode='zeros',
                 data_bits=8, parameter_bits=8, mode=QuantMode.QValue, o_bits=8,
                 clamp_data=None, clamp_weight=None, clamp_bias=None, ahead_relu=False, ahead_sigmoid=False):
        # venus limits
        if type(kernel_size) == int:
            assert kernel_size in (
                1, 2, 3, 4, 5), f"in Conv2dInt op, kernel size must be 1/2/3/4/5, but you have kernel size {kernel_size}"
        elif type(kernel_size) == tuple:
            assert kernel_size[0] in (1, 2, 3, 4, 5) and kernel_size[1] in (
                1, 2, 3, 4, 5), "in Conv2dInt op, kernel size must be 1/2/3/4/5, but you have kernel size {kernel_size}"

        if type(stride) == int:
            assert stride in (
                1, 2, 4), "in Conv2dInt op, stride size must be 1/2/4, but you have stride size {stride}"
        elif type(stride) == tuple:
            assert stride[0] in (1, 2, 4) and stride[1] in (
                1, 2, 4), "in Conv2dInt op, stride size must be 1/2/4, but you have stride size {stride}"

        if type(padding) == int:
            assert padding in (
                0, 1, 2, 3, 4), "in Conv2dInt op, padding size must be 1/2/4, but you have padding size {padding}"
        elif type(padding) == tuple:
            assert padding[0] in (0, 1, 2, 3, 4) and padding[1] in (
                0, 1, 2, 3, 4), "in Conv2dInt op, padding size must be 1/2/3/4/5, but you have padding size {padding}"

        if type(kernel_size) == int:
            kernel_size_h = kernel_size
            kernel_size_w = kernel_size
        elif type(kernel_size) == tuple:
            kernel_size_h = kernel_size[0]
            kernel_size_w = kernel_size[1]
        else:
            assert False, "kernel size type error."
        # if (groups != in_channels):
        #     assert math.ceil(in_channels/8) * 8 * kernel_size_h * kernel_size_w * math.ceil(out_channels/2) * 2 <= 32 * \
        #         1024, f"in Conv2dInt op, kernel must meet the requirements of non-depthwise convolution, but you have math.ceil({in_channels}/8) * 8 * {kernel_size_h} * {kernel_size_w} * math.ceil({out_channels}/2) * 2 <= 32 * 1024"
        # if (groups == in_channels):
        #     assert math.ceil(in_channels/16) * 16 * kernel_size_h * kernel_size_w <= 32 * \
        #         1024, f"in Conv2dInt op, kernel must meet the requirements of depthwise convolution, but you have math.ceil({in_channels}/16) * 16 * {kernel_size_h} * {kernel_size_w} <= 32 * 1024"

        if type(stride) == int:
            stride_h = stride
            stride_w = stride
        elif type(stride) == tuple:
            stride_h = stride[0]
            stride_w = stride[1]
        else:
            assert False, "kernel size type error."

        if type(padding) == int:
            padding_h = padding
            padding_w = padding
        elif type(padding) == tuple:
            padding_h = padding[0]
            padding_w = padding[1]
        else:
            assert False, "kernel size type error."

        assert kernel_size_h >= stride_h and kernel_size_w >= stride_w, f"kernel_size_h >= stride_h and kernel_size_w >= stride_w, but you have {kernel_size_h} < {stride_h} or {kernel_size_w} < {stride_w}"
        assert padding_h < kernel_size_h and padding_w < kernel_size_w, f"pad_h < weight_h && pad_w < weight_w, but you have {padding_h} >= {kernel_size_h} or {padding_w} >= {kernel_size_w}"

        nn.Conv2d.__init__(self, in_channels, out_channels,
                           kernel_size, stride, padding, dilation, groups, bias)
        ModuleIntConfig.__init__(
            self, data_bits=data_bits, parameter_bits=parameter_bits, mode=mode, o_bits=o_bits)
        self.params = torch.Tensor([self.stride[0], self.stride[1], self.padding[0],
                                   self.padding[1], self.dilation[0], self.dilation[1], groups])
        self.momentum = 0.1
        self.prefix = ""
        self.dump = False
        self.path = ""
        self.ahead_relu = ahead_relu
        self.ahead_sigmoid = ahead_sigmoid
        self.is_not_from_iqtensor = True
        self.clamp_data = clamp_data
        self.clamp_weight = clamp_weight
        self.clamp_bias = clamp_bias
        self.mode = mode
        self.register_buffer('running_x', torch.zeros(1))
        self.register_buffer('running_w', torch.zeros(1))
        self.register_buffer('running_o', torch.zeros(1))
        self.register_buffer('scale_x', torch.zeros(1))
        self.register_buffer('scale_w', torch.zeros(1))
        self.register_buffer('scale_o', torch.zeros(1))

    def forward(self, input):
        scale_x = ScalerBuffer(self.scale_x)
        running_x = ScalerBuffer(self.running_x)
        if isinstance(input, IQTensor):
            self.is_not_from_iqtensor = False
            if input.bits != self.data_bits:
                input = Requant.apply(
                    input, input.bits, input.scale_data, self.data_bits, self.mode)
            scale_x = ScalerBuffer(input.scale_data)
            running_x = ScalerBuffer(input.running_data)
        running_w = ScalerBuffer(self.running_w)
        running_o = ScalerBuffer(self.running_o)
        scale_w = ScalerBuffer(self.scale_w)
        scale_o = ScalerBuffer(self.scale_o)
        weight = self.weight
        bias = self.bias
        if weight.dtype == torch.float32:
            weight = normalize_weight_with_config(
                self.weight, self.clamp_weight, self.training)
            if self.bias is not None:
                bias = normalize_bias_with_config(
                    self.bias, self.clamp_bias, self.training)
        ret = ConvFunction.apply(input, weight, bias,
                                 self.kernel_size, self.padding, self.stride, self.dilation, self.groups,
                                 self.params, self.data_bits, self.parameter_bits, self.training, self.momentum,
                                 running_x, running_w, running_o, scale_x, scale_w, scale_o,
                                 self.prefix, self.dump, self.path, self.quant_mode, self.o_bits, self.quant, self.ahead_relu, self.is_not_from_iqtensor,
                                 self.clamp_data, self.clamp_weight, self.clamp_bias, self.ahead_sigmoid)

        self.running_x.fill_(running_x())
        self.running_w.fill_(running_w())
        self.running_o.fill_(running_o())
        self.scale_x.fill_(scale_x())
        self.scale_w.fill_(scale_w())
        self.scale_o.fill_(scale_o())
        return ret

    def state_dict(self, destination=None, prefix='', keep_vars=False):
        return ModuleIntConfig.state_dict_global(self, destination, prefix, keep_vars)

    def _load_from_state_dict(self, state_dict, prefix, local_metadata, strict,
                              missing_keys, unexpected_keys, error_msgs):
        ModuleIntConfig._load_from_state_dict_global(
            self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs)

    def extra_repr(self):
        s = nn.Conv2d.extra_repr(self)
        extra_s = ',clamp_data:{clamp_data},clamp_weight:{clamp_weight},clamp_bias:{clamp_bias},ahead_relu:{ahead_relu},ahead_sigmoid:{ahead_sigmoid}'.format(
            **self.__dict__)
        extra_s += ',data_bits:{data_bits},parameter_bits:{parameter_bits},o_bits:{o_bits}'.format(
            **self.__dict__)
        return s+extra_s



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
import itertools
from collections import OrderedDict

import torch
import torch.nn as nn
from torch.onnx import is_in_onnx_export

from ..config import config
from ..quant import (Quant, normalize_bias_with_config,
                     normalize_data_with_config, normalize_weight_with_config)
from ..utils import Dump, PlatFormQuant, QuantMode, ScalerBuffer
from .iqtensor import (IQTensor, from_torch_tensor, platform_to_string,
                       quantlinear)
from .ops import ModuleIntConfig
from .requant import Requant


class BatchNormFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input, running_mean, running_var, alpha, beta,
                training, exponential_average_factor, momentum, eps,
                running_x, running_w, running_o,
                eval_scale_x, eval_scale_w, eval_scale_o,
                data_bits, parameter_bits, prefix, dump, path, mode, o_bits, quant,
                is_not_from_iqtensor, ahead_relu, clamp_data, clamp_weight, clamp_bias):
        scale_x = None
        scale_w = None
        scale_o = None
        if training:
            ctx.clamp_data = clamp_data
            ctx.clamp_weight = clamp_weight
            ctx.clamp_bias = clamp_bias
            ctx.bits = data_bits, parameter_bits, o_bits
            zero_point = input.zero_point if isinstance(input, IQTensor) else 0
            is_iq_tensor = True if isinstance(input, IQTensor) else False
            ctx.value = zero_point, is_iq_tensor
            saved_tensors = [input, alpha, beta]

            q_alpha, scale_w, max_value_w = quant.quant(
                alpha, parameter_bits, mode=mode, quant_data='weight')

            scale_w = ScalerBuffer(scale_w)
            running_w.mul_(1-momentum).add_(momentum*max_value_w)
            ctx.eps = eps
            # mul-add mul
            if isinstance(input, IQTensor):
                q_input, _, max_value_x = quant.quant(
                    input.data, data_bits, eval_scale_x, mode=QuantMode.QValue, quant_data='input',
                    iq_zero_point=input.zero_point)
                scale_x = eval_scale_x
            else:
                q_input, scale_x, max_value_x = quant.quant(
                    input.data, data_bits, mode=mode, quant_data='input')
                scale_x = ScalerBuffer(scale_x)
                running_x.mul_(1-momentum).add_(momentum*max_value_x)
            q_input = q_input.float() if data_bits + parameter_bits <= 16 else q_input.double()
            q_alpha = q_alpha.float() if data_bits + parameter_bits <= 16 else q_alpha.double()
            # q_output = None
            q_outputs = q_input * q_alpha
            # mul-add add
            if config.PlatFormQuant.platform_quant in (PlatFormQuant.luna_quant,):
                q_beta = (beta * scale_w * scale_x + 0.5).floor()
                if data_bits + parameter_bits <= 16:
                    q_beta = q_beta.float().round()
                else:
                    q_beta = q_beta.double()
                q_outputs = q_outputs + q_beta
                outputs = quant.dequant(q_outputs, scale_x*scale_w)
            else:
                assert False, "linger only support luna quant."
            out_tensor = outputs
            scale_o = None
            if o_bits is not None:
                outputs = normalize_data_with_config(outputs, clamp_data)
                out_tensor = outputs
                q_outputs, scale_o, max_value_o = quant.quant(
                    outputs, o_bits, mode=mode, quant_data='output', ahead_relu=ahead_relu)
                scale_o = ScalerBuffer(scale_o)
                running_o.mul_(1-momentum).add_(momentum*max_value_o)
                outputs = quant.dequant(q_outputs, scale_o)
            saved_tensors += [out_tensor]
            ctx.scale = scale_x, scale_w, scale_o
            ctx.save_for_backward(*saved_tensors)

        else:
            assert running_x > 0, 'invalid running_x <= 0, please finetune training before eval'
            # assert config.PlatFormQuant.platform_quant in (PlatFormQuant.luna_quant,), 'linger only support luna quant.'
            if alpha.dtype == torch.float32:
                scale_x = ScalerBuffer(quant.running_to_scale(
                    running_x, data_bits, mode=mode))
                if o_bits is not None:
                    scale_o = ScalerBuffer(quant.running_to_scale(
                        running_o, o_bits, mode=mode))
                q_alpha, scale_w, _ = quant.quant(
                    alpha, parameter_bits, mode=mode, quant_data='weight')
                scale_w = ScalerBuffer(scale_w)
                q_beta = None
                if config.PlatFormQuant.platform_quant in (PlatFormQuant.luna_quant,):
                    q_beta = (beta * scale_w * scale_x + 0.5).floor()
                    if data_bits + parameter_bits <= 16:
                        q_beta = q_beta.float().round().double()
                else:
                    assert False, 'linger only support luna quant.'

            else:
                scale_x = eval_scale_x
                scale_w = eval_scale_w
                scale_o = eval_scale_o
                q_alpha = alpha.view(1, -1, 1, 1)
                q_beta = beta.view(1, -1, 1, 1).double()

            if isinstance(input, IQTensor):
                scale_x = eval_scale_x
                q_input, _, _ = quant.quant(
                    input.data, data_bits, scale_x, mode=QuantMode.QValue, quant_data='input', iq_zero_point=input.zero_point)
            else:
                q_input, _, _ = quant.quant(
                    input.data, data_bits, scale_x, mode=mode, quant_data='input')

            q_input = q_input.float() if data_bits + parameter_bits <= 16 else q_input.double()
            q_alpha = q_alpha.float() if data_bits + parameter_bits <= 16 else q_alpha.double()
            q_outputs = q_input * q_alpha + q_beta
            outputs = quant.dequant(q_outputs, scale_x*scale_w)
            if o_bits is not None:
                if config.PlatFormQuant.platform_quant == PlatFormQuant.luna_quant:
                    q_outputs, _, _ = quant.quant(
                        outputs, o_bits, scale_o, mode=mode, quant_data='output')
                else:
                    assert False, "linger only support luna quant."
                outputs = quant.dequant(q_outputs, scale_o)

            if dump:
                name_list = ["input", "running_mean", "running_var", "q_alpha", "q_input", "q_beta", "q_outputs",
                             "scale_x", "scle_w", "scale_o"]
                attr_list = [input, running_mean, running_var, q_alpha, q_input, q_beta, q_outputs,
                             scale_x.data, scale_w.data, scale_o.data]
                Dump.dump_file(prefix, ".BatchNormInt.",
                               zip(name_list, attr_list), path)

        if isinstance(scale_o, float):
            return from_torch_tensor(outputs, scale_o, o_bits)
        elif isinstance(scale_o, torch.Tensor):
            return from_torch_tensor(outputs, scale_o.item(), o_bits)
        else:
            return from_torch_tensor(outputs, scale_o.data, o_bits)

    @staticmethod
    def backward(ctx, gradOutput):
        input, alpha, beta, outputs = ctx.saved_tensors
        clamp_data = ctx.clamp_data
        data_bits, parameter_bits, o_bits = ctx.bits
        scale_x, scale_w, scale_o = ctx.scale
        zero_point, is_iq_tensor = ctx.value

        if is_iq_tensor:
            f_input = input.data
        else:
            q_input, _, _ = Quant.quant(
                input.data, data_bits, scale_x, mode=QuantMode.QValue, quant_data='input')
            f_input = Quant.dequant(q_input, scale_x)
        f_input = f_input.detach().clone().requires_grad_(True)
        q_alpha, _, _ = Quant.quant(
            alpha.data, parameter_bits, scale_w, mode=QuantMode.QValue, quant_data='weight')
        f_alpha = Quant.dequant(q_alpha, scale_w)
        f_alpha = f_alpha.detach().clone().requires_grad_(True)
        with torch.enable_grad():
            z = f_alpha * f_input + beta
            z = normalize_data_with_config(z, clamp_data)
            gradInput, gradAlpha, gradBeta = torch.autograd.grad(
                z, (f_input, f_alpha, beta), gradOutput)

        return gradInput, None, None, gradAlpha, gradBeta, \
            None, None, None, None, \
            None, None, None, None, \
            None, None, None, \
            None, None, None, \
            None, None, None, None, None, None, None, None, \
            None, None

    @staticmethod
    def symbolic(g, input, running_mean, running_var, weights, bias,
                 training, exponential_average_factor, momentum, eps,
                 running_x, running_w, running_o,
                 scale_x, scale_w, scale_o,
                 data_bits, parameter_bits, prefix, dump, path, mode, o_bits, quant,
                 is_not_from_iqtensor, ahead_relu, clamp_data, clamp_weight, clamp_bias):
        op_inner = None
        if is_not_from_iqtensor:
            platform_quant = platform_to_string(
                config.PlatFormQuant.platform_quant)
            op_inner = quantlinear(
                g, input, scale_x(), platform_quant, data_bits)
        # if config.PlatFormQuant.platform_quant in (PlatFormQuant.luna_quant,):
        param_dict = {'data_bits_i': data_bits, 'parameter_bits_i': parameter_bits, 'o_bits_i': o_bits,
                      'scale_x_f': scale_x(), 'scale_w_f': scale_w(), 'scale_o_f': scale_o()}

        platform_quant = platform_to_string(
            config.PlatFormQuant.platform_quant)
        param_dict['platform_quant_s'] = platform_quant
        if is_not_from_iqtensor:
            input_list = [op_inner, weights, bias, running_mean, running_var]
        else:
            input_list = [input, weights, bias, running_mean, running_var]
        return g.op("thinker::BatchNorm2dInt", *input_list, **param_dict)


class BatchNormInt(nn.BatchNorm2d, ModuleIntConfig):
    r"""实现BatchNormInt的量化训练与测试，继承自nn.BatchNorm2d,

    Args:
        num_features eps momentum affine track_running_stats
        标准nn.BatchNorm2d的参数
        data_bits(int): 输入量化位数
        parameter_bits(int): 参数量化位数
        mode(Enum): 量化方式，支持MaxValue与Qvalue
        o_bits(int, default=None):输出量化位数
        clamp_data(float or None): 针对输出的clamp数值
        clamp_weight(float or None): 针对转为乘加操作之后的weight与bias的clamp数值，此处不使用
        clamp_bias(float or None): 与clamp_weight一致
        ahead_relu(bool): 是否做融合relu之后的数值统计scale

    Examples:
        test/test_batchnorm_int.py

    """

    def __init__(self, num_features, eps=1e-05, momentum=0.1, affine=True, track_running_stats=True, data_bits=8, parameter_bits=8, mode=QuantMode.QValue, o_bits=None,
                 clamp_data=None, clamp_weight=None, clamp_bias=None, ahead_relu=False):
        nn.BatchNorm2d.__init__(self, num_features, eps,
                                momentum, affine, track_running_stats)
        ModuleIntConfig.__init__(
            self, data_bits=data_bits, parameter_bits=parameter_bits, mode=mode, o_bits=o_bits)
        self.prefix = ""
        self.dump = False
        self.path = ""
        self.is_not_from_iqtensor = True
        self.ahead_relu = ahead_relu
        self.clamp_data = clamp_data
        self.clamp_weight = clamp_weight
        self.clamp_bias = clamp_bias
        self.mode = mode
        self.register_buffer('running_x', torch.zeros(1))
        self.register_buffer('running_w', torch.zeros(1))
        self.register_buffer('running_o', torch.zeros(1))
        self.register_buffer('scale_x', torch.zeros(1))
        self.register_buffer('scale_w', torch.zeros(1))
        self.register_buffer('scale_o', torch.zeros(1))

    def forward(self, input):
        # assert (self.running_mean.abs().sum() != 0), 'batchnormint onlu support finetune'
        if self.momentum is None:
            exponential_average_factor = 0.0
        else:
            exponential_average_factor = self.momentum
        scale_x = ScalerBuffer(self.scale_x)
        running_x = ScalerBuffer(self.running_x)
        if isinstance(input, IQTensor):
            self.is_not_from_iqtensor = False
            if input.bits != self.data_bits:
                input = Requant.apply(
                    input, input.bits, input.scale_data, self.data_bits, self.mode)
            scale_x = ScalerBuffer(input.scale_data)
            running_x = ScalerBuffer(input.running_data)
        running_w = ScalerBuffer(self.running_w)
        running_o = ScalerBuffer(self.running_o)
        scale_w = ScalerBuffer(self.scale_w)
        scale_o = ScalerBuffer(self.scale_o)
        if self.training and self.track_running_stats:
            if self.num_batches_tracked is not None:
                self.num_batches_tracked = self.num_batches_tracked + 1
                if self.momentum is None:  # use cumulative moving average
                    exponential_average_factor = 1.0 / \
                        float(self.num_batches_tracked)
                else:  # use exponential moving average
                    exponential_average_factor = self.momentum
        momentum = 0.1
        # do fuse
        batchsize, channels, height, width = input.shape
        size = batchsize * height * width
        if self.training:
            mean = input.clone().sum((0, 2, 3), keepdim=True) / size
            var = input.clone().pow(2).sum((0, 2, 3), keepdim=True) / size - \
                (input.clone().sum((0, 2, 3), keepdim=True) / size).pow(2)
            var = torch.clamp(var, min=0.0)
            self.running_mean = (
                1 - self.momentum) * self.running_mean + self.momentum * mean.squeeze().detach()
            self.running_var = (1 - self.momentum) * self.running_var + \
                self.momentum * var.squeeze().detach()
        else:
            mean = self.running_mean.reshape(1, -1, 1, 1)
            var = self.running_var.reshape(1, -1, 1, 1)
        sigma = 1/torch.sqrt(var + self.eps)
        if self.weight.dtype == torch.float32:
            alpha = self.weight.view(1, -1, 1, 1)*sigma
            beta = self.bias.view(1, -1, 1, 1)-mean*alpha
            alpha = normalize_weight_with_config(
                alpha, self.clamp_weight, self.training)
            beta = normalize_bias_with_config(
                beta, self.clamp_bias, self.training)
        else:
            alpha = self.weight
            beta = self.bias

        ret = BatchNormFunction.apply(input, self.running_mean, self.running_var, alpha, beta,
                                      self.training or not self.track_running_stats, exponential_average_factor, momentum, self.eps,
                                      running_x, running_w, running_o, scale_x, scale_w, scale_o,
                                      self.data_bits, self.parameter_bits, self.prefix, self.dump, self.path, self.quant_mode, self.o_bits, self.quant, self.is_not_from_iqtensor, self.ahead_relu,
                                      self.clamp_data, self.clamp_weight, self.clamp_bias)

        self.running_x.fill_(running_x())
        self.running_w.fill_(running_w())
        self.running_o.fill_(running_o())

        self.scale_x.fill_(scale_x())
        self.scale_w.fill_(scale_w())
        self.scale_o.fill_(scale_o())
        return ret

    def state_dict(module, destination=None, prefix='', keep_vars=False):
        if destination is None:
            destination = OrderedDict()
            destination._metadata = OrderedDict()
        destination._metadata[prefix[:-1]
                              ] = local_metadata = dict(version=module._version)
        if is_in_onnx_export():
            assert module._buffers['running_x'] > 0, 'invalid running_x and running_o, please finetune first'
            scale_x = ScalerBuffer(module._buffers['scale_x'])
            if module.is_not_from_iqtensor:
                scale_x = ScalerBuffer(module.quant.running_to_scale(ScalerBuffer(
                    module._buffers['running_x']), module.data_bits, mode=module.quant_mode))
                module._buffers['scale_x'].data.fill_(scale_x())

            if module.o_bits is not None:
                scale_o = ScalerBuffer(module.quant.running_to_scale(ScalerBuffer(
                    module._buffers['running_o']), module.o_bits, mode=module.quant_mode))
                module._buffers['scale_o'].data.fill_(scale_o())

        if 'scale_w' in module._buffers and module._parameters['weight'].dtype == torch.float:
            weights = module._parameters['weight'].data
            bias = module._parameters['bias'].data
            mean = module._buffers['running_mean'].data.view(1, -1, 1, 1)
            var = module._buffers['running_var'].data.view(1, -1, 1, 1)
            sigma = 1/torch.sqrt(var + module.eps)
            alpha = weights.view(1, -1, 1, 1)*sigma
            beta = bias.view(1, -1, 1, 1)-mean*alpha
            alpha = normalize_weight_with_config(
                alpha, module.clamp_weight, False)
            beta = normalize_bias_with_config(beta, module.clamp_bias, False)
            q_alpha = None
            q_beta = None
            if is_in_onnx_export():
                q_alpha, scale_w, _ = module.quant.quant(
                    alpha, module.parameter_bits, mode=module.quant_mode)
                scale_w = ScalerBuffer(scale_w)
                module._buffers['scale_w'].data.fill_(scale_w())
                if config.PlatFormQuant.platform_quant in (PlatFormQuant.luna_quant,):
                    q_beta = (beta * scale_w * scale_x +
                              0.5).floor().float().round().int()
                else:
                    assert False, 'linger only support luna quant.'

            weight_tensor = module._parameters['weight']
            bias_tensor = module._parameters['bias']
            if is_in_onnx_export():
                if module.parameter_bits <= 8:
                    weight_tensor.data = q_alpha.char().reshape(-1)
                    weight_tensor.char()
                elif module.parameter_bits <= 16:
                    weight_tensor.data = q_alpha.short().reshape(-1)
                    weight_tensor.short()
                else:
                    weight_tensor.data = q_alpha.int().reshape(-1)
                    weight_tensor.int()
                bias_tensor.data = q_beta.int().reshape(-1)
                bias_tensor.int()
        module._save_to_state_dict(destination, prefix, keep_vars)
        for name, module in module._modules.items():
            if module is not None:
                module.state_dict(destination, prefix +
                                  name + '.', keep_vars=keep_vars)
        for hook in module._state_dict_hooks.values():
            hook_result = hook(module, destination, prefix, local_metadata)
            if hook_result is not None:
                destination = hook_result
        return destination

    def _load_from_state_dict(module, state_dict, prefix, local_metadata, strict,
                              missing_keys, unexpected_keys, error_msgs):
        allow_missing_keys = ['running_w', 'running_x', 'running_o',
                              'scale_x', 'scale_w', 'scale_o']
        local_missing_keys = []
        module._load_from_state_dict_global_(state_dict, prefix, local_metadata, strict,
                                             local_missing_keys, unexpected_keys, error_msgs)
        matched = True
        fake_missing_keys = []
        for k_local in local_missing_keys:
            if k_local.replace(prefix, '', 1) not in allow_missing_keys:
                matched = False
                fake_missing_keys.append(k_local)
        if matched:
            local_missing_keys = []
        else:
            local_missing_keys = fake_missing_keys
        missing_keys += local_missing_keys

    def _load_from_state_dict_global_(module, state_dict, prefix, local_metadata, strict,
                                      missing_keys, unexpected_keys, error_msgs):
        for hook in module._load_state_dict_pre_hooks.values():
            hook(state_dict, prefix, local_metadata, strict,
                 missing_keys, unexpected_keys, error_msgs)
        local_name_params = itertools.chain(
            module._parameters.items(), module._buffers.items())
        local_state = {k: v.data for k,
                       v in local_name_params if v is not None}
        for name, param in local_state.items():
            key = prefix + name
            if key in state_dict:
                input_param = state_dict[key]

                if len(param.shape) == 0 and len(input_param.shape) == 1:
                    input_param = input_param[0]

                if input_param.shape != param.shape:
                    error_msgs.append('size mismatch for {}: copying a param with shape {} from checkpoint, '
                                      'the shape in current model is {}.'
                                      .format(key, input_param.shape, param.shape))
                    continue

                if isinstance(input_param, torch.nn.Parameter):
                    input_param = input_param.data
                try:
                    param.copy_(input_param)
                    if input_param.dtype == torch.int32 or input_param.dtype == torch.int8 or input_param.dtype == torch.int16:
                        module._parameters[name] = param.int()

                except Exception:
                    error_msgs.append('While copying the parameter named "{}", '
                                      'whose dimensions in the model are {} and '
                                      'whose dimensions in the checkpoint are {}.'
                                      .format(key, param.size(), input_param.size()))
            elif strict:
                missing_keys.append(key)
        if strict:
            for key in state_dict.keys():
                if key.startswith(prefix):
                    input_name = key[len(prefix):]
                    input_name = input_name.split('.', 1)[0]
                    if input_name not in module._modules and input_name not in local_state:
                        unexpected_keys.append(key)

    def extra_repr(self):
        s = nn.BatchNorm2d.extra_repr(self)
        extra_s = ' ,clamp_data:{clamp_data},clamp_weight:{clamp_weight},clamp_bias:{clamp_bias},ahead_relu:{ahead_relu}'.format(
            **self.__dict__)
        extra_s += ',data_bits:{data_bits}, parameter_bits:{parameter_bits}, o_bits:{o_bits}'.format(
            **self.__dict__)
        return s+extra_s



import math

import torch

from ..utils import QuantMode
from .iqtensor import IQTensor, from_torch_tensor


class Requant(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input, bits_src, scale_src, bits_dst, mode=QuantMode.QValue):
        s_rescale = (math.pow(2, bits_dst-1) - 1.0) / \
            (math.pow(2, bits_src-1) - 1.0)
        if mode == QuantMode.QValue:
            s_rescale = math.pow(2, round(math.log(s_rescale, 2)))
        scale_bst = s_rescale*scale_src
        zero_point = 0
        if isinstance(input, IQTensor):
            zero_point = input.zero_point
        if zero_point != 0:
            zero_point = math.pow(2, bits_dst-1)
        s = from_torch_tensor(input, scale_bst, bits_dst, zero_point=zero_point)
        s.requant_()
        return s

    @staticmethod
    def backward(ctx, gradOutput):
        return gradOutput, None, None, None, None

    @staticmethod
    def symbolic(g, input, bits_src, scale_src, bits_dst, mode=QuantMode.QValue):
        s_rescale = (math.pow(2, bits_dst-1) - 1.0) / \
            (math.pow(2, bits_src-1) - 1.0)
        if mode == QuantMode.QValue:
            s_rescale = math.pow(2, round(math.log(s_rescale, 2)))
        scale_bst = s_rescale*scale_src
        return g.op("thinker::Requant", input, data_bits_i=bits_src, scale_x_f=scale_src, scale_o_f=scale_bst, o_bits_i=bits_dst)


__all__ = ['Requant']



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

def _to_int(s):
    try:
        return int(s)
    except ValueError:
        return s


__version__ = "0.9.0"
version_info = tuple(_to_int(s) for s in __version__.split("."))



import linger
import pytest
import torch
import torch.nn as nn
from linger.ops import *


def test_linear():
    torch.manual_seed(1)
    torch.cuda.manual_seed(1)
    origin_fc = nn.Linear(50, 100, True).cuda().train()
    linger.SetPlatFormQuant(platform_quant=linger.PlatFormQuant.luna_quant)
    my_fc = LinearInt(50, 100, True, data_bits=8,
                      parameter_bits=8).cuda().train()
    weight = torch.randn(100, 50)
    bias = torch.randn(100)
    origin_fc.weight.data.copy_(weight)
    origin_fc.bias.data.copy_(bias)
    my_fc.weight.data.copy_(weight)
    my_fc.bias.data.copy_(bias)
    input1 = torch.ones(50, 50, requires_grad=True).cuda()
    for epoch in range(10):
        output1 = origin_fc(input1)
        output2 = my_fc(input1)
        criterion = nn.MSELoss()
        optimizer = torch.optim.SGD(origin_fc.parameters(), lr=0.1)
        target = torch.ones(50, 100).cuda()
        loss = criterion(output1, target)
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        optimizer1 = torch.optim.SGD(my_fc.parameters(), lr=0.1)
        loss1 = criterion(output2, target)
        optimizer1.zero_grad()
        loss1.backward()
        optimizer1.step()
        print('loss1, loss2: ', loss, loss1)
        assert criterion(origin_fc.weight, my_fc.weight) < 0.02


def test_conv():
    torch.manual_seed(1)
    origin_conv = nn.Conv2d(10, 10, 1, 1, 0, 1, 1).cuda()
    linger.SetPlatFormQuant(platform_quant=linger.PlatFormQuant.luna_quant)
    my_conv = Conv2dInt(10, 10, 1, 1, 0, 1, 1, data_bits=8,
                        parameter_bits=8, o_bits=8).cuda()
    weight = torch.randn(10, 10, 1, 1)
    bias = torch.randn(10)
    origin_conv.weight.data.copy_(weight)
    origin_conv.bias.data.copy_(bias)
    my_conv.weight.data.copy_(weight)
    my_conv.bias.data.copy_(bias)
    input1 = torch.ones(10, 10, 50, 50, requires_grad=True).cuda()
    for epoch in range(10):
        output1 = origin_conv(input1)
        output2 = my_conv(input1)
        criterion = nn.MSELoss()
        optimizer = torch.optim.SGD(origin_conv.parameters(), lr=0.01)
        target = torch.ones(10, 10, 50, 50, requires_grad=True).cuda()
        loss = criterion(output1, target)
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        optimizer1 = torch.optim.SGD(my_conv.parameters(), lr=0.01)
        loss1 = criterion(output2, target)
        optimizer1.zero_grad()
        loss1.backward()
        optimizer1.step()
        assert criterion(origin_conv.weight, my_conv.weight) < 0.02


class bn(nn.Module):
    def __init__(self):
        super(bn, self).__init__()

        self.out = nn.BatchNorm2d(100)

    def forward(self, input):
        return self.out(input)


class bn_(nn.Module):
    def __init__(self):
        super(bn_, self).__init__()

        self.out = BatchNormInt(100)

    def forward(self, input):
        return self.out(input)


def test_convtranspose():
    torch.manual_seed(1)
    torch.cuda.manual_seed(1)
    origin_convtranspose = nn.ConvTranspose2d(
        10, 10, 5, 5, 2, 4, 2, True, 2).cuda().train()
    my_convtranspose = ConvTranspose2dInt(
        10, 10, 5, 5, 2, 4, 2, True, 2, data_bits=16, parameter_bits=16, o_bits=8).cuda().train()
    weight = torch.randn(10, 5, 5, 5)
    bias = torch.randn(10)
    origin_convtranspose.weight.data.copy_(weight)
    origin_convtranspose.bias.data.copy_(bias)
    my_convtranspose.weight.data.copy_(weight)
    my_convtranspose.bias.data.copy_(bias)
    input1 = torch.randn(10, 10, 50, 50, requires_grad=True).cuda()
    for epoch in range(10):
        output1 = origin_convtranspose(input1)
        output2 = my_convtranspose(input1)
        criterion = nn.MSELoss()
        optimizer = torch.optim.SGD(origin_convtranspose.parameters(), lr=0.01)
        target = torch.ones(10, 10, 254, 254, requires_grad=True).cuda()
        loss = criterion(output1, target)
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        optimizer1 = torch.optim.SGD(my_convtranspose.parameters(), lr=0.01)
        loss1 = criterion(output2, target)
        optimizer1.zero_grad()
        loss1.backward()
        optimizer1.step()
        print("loss1: {}, loss2: {}".format(loss, loss1))
        assert criterion(origin_convtranspose.weight,
                         my_convtranspose.weight) < 0.02


def test_batchnorm():
    torch.manual_seed(1)
    torch.cuda.manual_seed(1)
    origin_bn = nn.BatchNorm2d(10, eps=1e-5, momentum=0.1).cuda().train()
    my_bn = BatchNormInt(10, eps=1e-5, momentum=0.1, o_bits=8,
                         data_bits=8, parameter_bits=8).cuda().train()
    weight = torch.randn(10)
    bias = torch.randn(10)
    running_mean = torch.randn(10)
    running_var = torch.randn(10)
    origin_bn.weight.data.copy_(weight)
    origin_bn.bias.data.copy_(bias)
    origin_bn.running_mean.data.copy_(running_mean)
    origin_bn.running_var.data.copy_(running_var)

    my_bn.weight.data.copy_(weight)
    my_bn.bias.data.copy_(bias)
    my_bn.running_mean.data.copy_(running_mean)
    my_bn.running_var.data.copy_(running_var)
    # target = torch.ones(10, 10, 50, 50, requires_grad=True).cuda()
    grad = torch.randn(10, 10, 100, 100).cuda()
    linger.SetPlatFormQuant(platform_quant=linger.PlatFormQuant.luna_quant)
    input1 = torch.randn(10, 10, 100, 100, requires_grad=True).cuda()
    input2 = input1.detach().clone()

    for i in range(100):
        output1 = origin_bn(input1)
        output2 = my_bn(input2)
    output1 = origin_bn.eval()(input1)
    output2 = my_bn.eval()(input2)
    assert (output1.abs().mean() - output2.abs().mean()).abs() < 1/127


def test_gru():
    torch.manual_seed(1)
    torch.cuda.manual_seed(1)
    origin_gru = nn.GRU(10, 20, batch_first=True,
                        bidirectional=True).cuda().train()
    my_gru = GRUInt(10, 20, batch_first=True,
                    bidirectional=True, o_bits=8).cuda().train()
    weight_ih = torch.randn(60, 10)
    weight_hh = torch.randn(60, 20)
    bias_ih = torch.randn(60)
    bias_hh = torch.randn(60)
    origin_gru.weight_ih_l0.data.copy_(weight_ih)
    origin_gru.weight_hh_l0.data.copy_(weight_hh)
    origin_gru.bias_ih_l0.data.copy_(bias_ih)
    origin_gru.bias_hh_l0.data.copy_(bias_hh)
    my_gru.weight_ih_l0.data.copy_(weight_ih)
    my_gru.weight_hh_l0.data.copy_(weight_hh)
    my_gru.bias_ih_l0.data.copy_(bias_ih)
    my_gru.bias_hh_l0.data.copy_(bias_hh)
    input1 = torch.ones(10, 5, 10, requires_grad=True).cuda()
    for epoch in range(10):
        output1, hy = origin_gru(input1)
        output2, hy = my_gru(input1)
        criterion = nn.MSELoss()
        optimizer = torch.optim.SGD(origin_gru.parameters(), lr=0.1)
        target = torch.ones(10, 5, 40).cuda()
        loss = criterion(output1, target)
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        optimizer1 = torch.optim.SGD(my_gru.parameters(), lr=0.1)
        loss1 = criterion(output2, target)
        optimizer1.zero_grad()
        loss1.backward()
        optimizer1.step()
        print('loss1, loss2: ', loss, loss1)
        assert criterion(origin_gru.weight_ih_l0, my_gru.weight_ih_l0) < 0.02
        assert criterion(origin_gru.weight_hh_l0, my_gru.weight_hh_l0) < 0.02
        assert criterion(origin_gru.bias_ih_l0, my_gru.bias_ih_l0) < 0.02
        assert criterion(origin_gru.bias_hh_l0, my_gru.bias_hh_l0) < 0.02


def test_lstm():
    torch.manual_seed(1)
    torch.cuda.manual_seed(1)
    origin_lstm = nn.LSTM(10, 20, batch_first=True,
                          bidirectional=True).cuda().train()
    my_lstm = LSTMInt(10, 20, batch_first=True,
                      bidirectional=True, o_bits=8).cuda().train()
    weight_ih = torch.randn(80, 10)
    weight_hh = torch.randn(80, 20)
    bias_ih = torch.randn(80)
    bias_hh = torch.randn(80)
    origin_lstm.weight_ih_l0.data.copy_(weight_ih)
    origin_lstm.weight_hh_l0.data.copy_(weight_hh)
    origin_lstm.bias_ih_l0.data.copy_(bias_ih)
    origin_lstm.bias_hh_l0.data.copy_(bias_hh)
    my_lstm.weight_ih_l0.data.copy_(weight_ih)
    my_lstm.weight_hh_l0.data.copy_(weight_hh)
    my_lstm.bias_ih_l0.data.copy_(bias_ih)
    my_lstm.bias_hh_l0.data.copy_(bias_hh)
    input1 = torch.ones(10, 5, 10, requires_grad=True).cuda()
    for epoch in range(10):
        output1, (hy, cy) = origin_lstm(input1)
        output2, (hy, cy) = my_lstm(input1)
        criterion = nn.MSELoss()
        optimizer = torch.optim.SGD(origin_lstm.parameters(), lr=0.1)
        target = torch.ones(10, 5, 40).cuda()
        loss = criterion(output1, target)
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        optimizer1 = torch.optim.SGD(my_lstm.parameters(), lr=0.1)
        loss1 = criterion(output2, target)
        optimizer1.zero_grad()
        loss1.backward()
        optimizer1.step()
        print('loss1, loss2: ', loss, loss1)
        assert criterion(origin_lstm.weight_ih_l0, my_lstm.weight_ih_l0) < 0.02
        assert criterion(origin_lstm.weight_hh_l0, my_lstm.weight_hh_l0) < 0.02
        assert criterion(origin_lstm.bias_ih_l0, my_lstm.bias_ih_l0) < 0.02
        assert criterion(origin_lstm.bias_hh_l0, my_lstm.bias_hh_l0) < 0.02



from typing import Dict, List, Set

import torch
import torch.nn

from ..ops.iqtensor import IQTensor, from_torch_tensor


class ScopedEnterFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx, t, scope):
        m = t.clone()
        if isinstance(t, IQTensor):
            m = from_torch_tensor(m, t.scale_data, t.bits, t.zero_point)
        return m

    @staticmethod
    def backward(ctx, gradOutput):
        return gradOutput, None

    @staticmethod
    def symbolic(g, input, scope):
        return g.op("thinker::ScopedEnter", input, scope_s=scope)


class ScopedLeaveFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx, t, scope):
        m = t.clone()
        if isinstance(t, IQTensor):
            m = from_torch_tensor(m, t.scale_data, t.bits, t.zero_point)
        return m

    @staticmethod
    def backward(ctx, gradOutput):
        return gradOutput, None

    @staticmethod
    def symbolic(g, input, scope):
        return g.op("thinker::ScopedLeave", input, scope_s=scope)


global_scoped = {}


def scope_forward(input, scope_name, func):
    list_r = []
    for t in input:
        if isinstance(t, torch.Tensor):
            list_r.append(func.apply(t, scope_name))
        elif isinstance(t, dict):
            dict_r = {}
            for k, v in t.items():
                if isinstance(v, torch.Tensor):
                    dict_r[k] = func.apply(v, scope_name)
                else:
                    dict_r[k] = v
            list_r.append(dict_r)
        elif isinstance(t, (tuple, list)):
            list_r.append(scope_forward(t, scope_name, func))
        else:
            list_r.append(t)
    return type(input)(list_r)


def enter_forward(module, input):
    scope_name = global_scoped[module][0]
    if isinstance(input, torch.Tensor):
        return ScopedEnterFunction.apply(input, scope_name)
    list_r = None
    if isinstance(input, (tuple, list)):
        list_r = scope_forward(input, scope_name, ScopedEnterFunction)
    elif isinstance(input, dict):
        dict_r = {}
        for k, v in input.items():
            assert isinstance(
                v, torch.Tensor), 'dict input only support depth=1'
            dict_r[k] = ScopedEnterFunction.apply(v, scope_name)
        return dict_r
    else:
        assert 0, 'foward only support tensor, tuple or dict of tensor'
    return type(input)(list_r)


def leave_forward(module, input, output):
    scope_name = global_scoped[module][0]
    if isinstance(output, torch.Tensor):
        return ScopedLeaveFunction.apply(output, scope_name)
    list_r = None
    if isinstance(output, (tuple, list)):
        list_r = scope_forward(output, scope_name, ScopedLeaveFunction)
    elif isinstance(output, dict):
        dict_r = {}
        for k, v in input.items():
            if isinstance(v, torch.Tensor):
                dict_r[k] = ScopedLeaveFunction.apply(v, scope_name)
            else:
                dict_r[k] = v
        return dict_r
    else:
        assert 0, 'foward only support tensor or tuple of tensor'
    return type(output)(list_r)


def build_module_scope(model: torch.nn.Module):
    scopes = {}
    for name, child in model.named_children():
        hook_handle_foward_pre = child.register_forward_pre_hook(enter_forward)
        hook_handle_forward = child.register_forward_hook(leave_forward)
        scopes[child] = (name, hook_handle_foward_pre, hook_handle_forward)
        scopes.update(build_module_scope(child))
    return scopes


def build_global_scope(model: torch.nn.Module):
    global global_scoped
    global_scoped = build_module_scope(model)
    hook_handle_foward_pre = model.register_forward_pre_hook(
        enter_forward)  # not execute before forward
    hook_handle_forward = model.register_forward_hook(leave_forward)
    global_scoped[model] = (".", hook_handle_foward_pre, hook_handle_forward)
    return global_scoped


def remove_scoped_node(model, op_type):
    graph_output_name = []
    for ii in model.graph.output:
        graph_output_name.append(ii.name)
    nodes = model.graph.node[::-1]
    for i, node in enumerate(nodes):
        if node.op_type == op_type:
            model.graph.node.remove(node)
            if node.output[0] in graph_output_name:
                for each in model.graph.node:
                    for idx in range(len(each.output)):
                        if each.output[idx] == node.input[0]:
                            each.output[idx] = node.output[0]
                for each in model.graph.node:
                    for idx in range(len(each.input)):
                        if each.input[idx] == node.input[0]:
                            each.input[idx] = node.output[0]

                for each in model.graph.node:
                    if each.op_type == "If":
                        for gi in range(len(each.attribute)):
                            for x_node in each.attribute[gi].g.node:
                                for idx in range(len(x_node.output)):
                                    if x_node.output[idx] == node.input[0]:
                                        x_node.output[idx] = node.output[0]
                                for idx in range(len(x_node.input)):
                                    if x_node.input[idx] == node.input[0]:
                                        x_node.input[idx] = node.output[0]

            else:
                for each in model.graph.node:
                    for idx in range(len(each.input)):
                        if each.input[idx] == node.output[0]:
                            each.input[idx] = node.input[0]
                for each in model.graph.node:
                    if each.op_type == "If":
                        for gi in range(len(each.attribute)):
                            for x_node in each.attribute[gi].g.node:
                                for idx in range(len(x_node.input)):
                                    if x_node.input[idx] == node.output[0]:
                                        x_node.input[idx] = node.input[0]
    return model


def build_onnx_scope_info(onnx_model):

    nodes = onnx_model.graph.node
    node_son_dict: Dict[int, List[str]] = {}  # 保存所有的nodename 和对应的子节点，深度遍历使用
    node_scope_dict: Dict[int, List[str]] = {}  # 遍历后的名字保存 保存节点名字和对应的域名scope
    visited_node_set: Set[int] = set()  # 保存已经访问的节点
    # 先建立node之间的邻接关系
    nodeoutdict: Dict[int, str] = {}
    for node in nodes:
        for nodeout in node.output:
            nodeoutdict[nodeout] = node
        node_son_dict[id(node)] = []
    for node in nodes:
        for nodein in node.input:
            nodefather = nodeoutdict.get(nodein)
            if nodefather is not None:
                node_son_dict[id(nodefather)].append(node)
    # 根据邻接关系深度遍历所有节点
    nodetree = []  # 深度遍历栈
    # 加入图的所有输入节点  连接在graph.input上的第一个节点  原始只考虑了单一输入
    initializer_names = []
    for x in onnx_model.graph.initializer:
        initializer_names.append(x.name)
    input_idxs = []
    for idx, y in enumerate(onnx_model.graph.input):
        if y.name not in initializer_names:
            input_idxs.append(y.name)
    for node in onnx_model.graph.node[::-1]:
        for node_inp in node.input[::-1]:
            if node_inp in input_idxs:
                nodetree.append(node)
                visited_node_set.add(id(node))
                continue
    while len(nodetree) > 0:
        node = nodetree.pop()  # 获得栈中的节点
        node_sons = node_son_dict[id(node)]  # 获得当前节点的所有子节点
        if id(node) not in node_scope_dict:
            node_scope_dict[id(node)] = []
        for node_son in node_sons:
            if id(node_son) in visited_node_set:
                continue
            scopelist = node_scope_dict[id(node)].copy()  # 每个子节点都要记录一份当前的域
            if node.op_type == "ScopedEnter":
                scope = node.attribute[0].s.decode('utf-8')
                scopelist.append(scope)
            elif node.op_type == "ScopedLeave":
                scope = node.attribute[0].s.decode('utf-8')
                scopelist.pop()
            node_scope_dict[id(node_son)] = scopelist
            nodetree.append(node_son)  # 子节点入栈
            visited_node_set.add(id(node_son))  # 访问过子节点加入已访问
        # 遍历子节点结束后，保存修改的name
        if node.op_type != "ScopedEnter" and node.op_type != "ScopedLeave":
            newnamelist = node_scope_dict[id(node)]
            if len(newnamelist) > 1 and newnamelist[0] != '.':
                node.name = "." + ".".join(newnamelist) + "." + node.name
            elif len(newnamelist) > 1 and newnamelist[0] == '.':
                node.name = "." + ".".join(newnamelist[1:]) + "." + node.name
            elif len(newnamelist) == 1 and newnamelist[0] != '.':
                node.name = "." + newnamelist[0] + "." + node.name
            elif len(newnamelist) == 1 and newnamelist[0] == '.':
                node.name = "." + node.name
    # 删除多余节点
    onnx_model = remove_scoped_node(onnx_model, "ScopedEnter")
    onnx_model = remove_scoped_node(onnx_model, "ScopedLeave")

    return onnx_model


__all__ = ['build_global_scope', 'build_onnx_scope_info']



from collections import OrderedDict
from typing import Optional

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor
from torch.onnx import is_in_onnx_export

from ..quant import normalize_data_with_config, normalize_weight_with_config
from ..utils import Dump, QuantMode, ScalerBuffer
from .iqtensor import IQTensor, from_torch_tensor
from .ops import ModuleIntConfig
from .requant import Requant
from ..quant import Quant


class EmbeddingFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input, num_embeddings, embedding_dim, padding_idx, max_norm, norm_type, scale_grad_by_freq, sparse,
                weight, data_bits, parameter_bits, training, momentum, running_x, running_w, running_o, eval_scale_x, eval_scale_w, eval_scale_o,
                prefix, dump, path, mode, o_bits, quant,  is_not_from_iqtensor, clamp_data, clamp_weight, clamp_bias):
        scale_o = None
        if training:
            ctx.clamp_data = clamp_data
            ctx.clamp_weight = clamp_weight
            ctx.clamp_bias = clamp_bias
            ctx.bits = o_bits, parameter_bits
            saved_tensors = [input, weight]
            ctx.params = [num_embeddings, embedding_dim, padding_idx,
                          max_norm, norm_type, scale_grad_by_freq, sparse, ]

            # weights = normalize_weight_with_config(weight, clamp_weight, False)
            weights = weight
            q_weights, scale_w, max_value_w = quant.quant(
                weights, parameter_bits, mode=mode, quant_data='weight')
            running_w.mul_(1-momentum).add_(momentum*max_value_w)
            q_weights = q_weights.float() if data_bits + \
                parameter_bits <= 16 else q_weights.double()
            q_outputs = F.embedding(
                input, q_weights, padding_idx, max_norm, norm_type, scale_grad_by_freq, sparse)

            ctx.save_for_backward(*saved_tensors)

            outputs = quant.dequant(q_outputs, scale_w)
            if o_bits is not None:
                running_o.fill_(running_w())
                scale_o = scale_w
            ctx.scale_w = scale_w

        else:
            scale_x = None
            scale_w = None
            scale_o = eval_scale_o
            q_weights = None
            if weight.dtype == torch.float32:
                scale_x = ScalerBuffer(quant.running_to_scale(
                    running_x, data_bits, mode=mode))
                weights = weight
                # weights = normalize_weight_with_config(
                #     weight, clamp_weight, False)
                q_weights, scale_w, _ = quant.quant(
                    weights, parameter_bits, mode=mode, quant_data='weight')
                scale_w = ScalerBuffer(scale_w)
                if o_bits is not None:
                    assert running_o > 0, 'invalid running_o <= 0, please finetune training'
                    scale_o = scale_w
            else:
                scale_x = eval_scale_x
                scale_w = eval_scale_w
                q_weights = weight.double()
                if o_bits is not None:
                    scale_o = eval_scale_w
            q_weights = q_weights.double()
            q_outputs = F.embedding(
                input, q_weights, padding_idx, max_norm, norm_type, scale_grad_by_freq, sparse)

            outputs = quant.dequant(q_outputs, scale_w)
            if dump:
                name_list = ["input", "outputs",  "q_outputs", "q_weights",
                             "scale_x", "scale_w" "scale_o", "running_x", "running_o"]
                attr_list = [input, outputs, q_outputs, q_weights, scale_x.data,
                             scale_w.data, scale_o.data, running_x.data,  running_o.data]
                Dump.dump_file(prefix, ".EmbeddingInt.",
                               zip(name_list, attr_list), path)

        if o_bits is None:
            return outputs
        elif isinstance(scale_o, float):
            return from_torch_tensor(outputs, scale_o, o_bits)
        elif isinstance(scale_o, torch.Tensor):
            return from_torch_tensor(outputs, scale_o.item(), o_bits)
        else:
            return from_torch_tensor(outputs, scale_o.data, o_bits)

    @staticmethod
    def backward(ctx, gradOutput):
        clamp_data = ctx.clamp_data
        o_bits, parameter_bits = ctx.bits
        scale_w = ctx.scale_w
        input, weight,  = ctx.saved_tensors
        num_embeddings, embedding_dim, padding_idx, max_norm, norm_type, scale_grad_by_freq, sparse, = ctx.params
        q_weights, _, _ = Quant.quant(
            weight.data, o_bits, scale_w, mode=QuantMode.QValue, quant_data='weight')
        f_weights = Quant.dequant(q_weights, scale_w)
        f_weights = f_weights.detach().clone().requires_grad_(True)
        with torch.enable_grad():
            # weights = normalize_weight_with_config(weight, clamp_weight, True)
            # weights = weight
            z = F.embedding(input, f_weights, padding_idx, max_norm,
                            norm_type, scale_grad_by_freq, sparse)
            if o_bits is not None:
                z = normalize_data_with_config(
                    z, clamp_data)
            gradWeight, = torch.autograd.grad(z, (f_weights), gradOutput)
        return None, None, None, None, None, None, None, None, gradWeight,\
            None, None, None, None, None, None, None, None, None, None,\
            None, None, None, None, None, None,  None, None, None, None, None

    @staticmethod
    def symbolic(g, input, num_embeddings, embedding_dim, padding_idx, max_norm, norm_type, scale_grad_by_freq, sparse,
                 weight, data_bits, parameter_bits, training, momentum, running_x, running_w, running_o, eval_scale_x, eval_scale_w, eval_scale_o,
                 prefix, dump, path, mode, o_bits, quant,  is_not_from_iqtensor, clamp_data, clamp_weight, clamp_bias):

        return g.op("thinker::Gather", weight, input, scale_w_f=eval_scale_w(), scale_o_f=eval_scale_o(), parameter_bits_i=parameter_bits, o_bits_i=o_bits)


class EmbeddingInt(nn.Embedding, ModuleIntConfig):
    def __init__(self, num_embeddings: int, embedding_dim: int, padding_idx: Optional[int] = None,
                 max_norm: Optional[float] = None, norm_type: float = 2., scale_grad_by_freq: bool = False,
                 sparse: bool = False, _weight: Optional[Tensor] = None,
                 data_bits=8, parameter_bits=8, mode=QuantMode.QValue, o_bits=None,
                 clamp_data=None, clamp_weight=None, clamp_bias=None):
        nn.Embedding.__init__(self, num_embeddings, embedding_dim, padding_idx,
                              max_norm, norm_type, scale_grad_by_freq, sparse, _weight)
        ModuleIntConfig.__init__(
            self, data_bits=data_bits, parameter_bits=parameter_bits, mode=mode, o_bits=o_bits)
        self.momentum = 0.1
        self.prefix = ""
        self.dump = False
        self.path = ""
        self.is_not_from_iqtensor = True
        self.clamp_data = clamp_data
        self.clamp_weight = clamp_weight
        self.clamp_bias = clamp_bias
        self.mode = mode
        self.register_buffer('running_x', torch.zeros(1))
        self.register_buffer('running_w', torch.zeros(1))
        self.register_buffer('running_o', torch.zeros(1))
        self.register_buffer('scale_x', torch.zeros(1))
        self.register_buffer('scale_w', torch.zeros(1))
        self.register_buffer('scale_o', torch.zeros(1))

    def forward(self, input):
        scale_x = ScalerBuffer(self.scale_x)
        running_x = ScalerBuffer(self.running_x)
        if isinstance(input, IQTensor):
            # assert False, "embeddingint   input shouldn't be IQTensor !"
            self.is_not_from_iqtensor = False
            if input.bits != self.data_bits:
                input = Requant.apply(
                    input, input.bits, input.scale_data, self.data_bits, self.mode)
            scale_x = ScalerBuffer(input.scale_data)
            running_x = ScalerBuffer(input.running_data)
        running_w = ScalerBuffer(self.running_w)
        running_o = ScalerBuffer(self.running_o)
        scale_w = ScalerBuffer(self.scale_w)
        scale_o = ScalerBuffer(self.scale_o)

        weights = self.weight
        if self.weight.dtype == torch.float32:
            weights = normalize_weight_with_config(
                self.weight, self.clamp_weight, self.training)

        ret = EmbeddingFunction.apply(input, self.num_embeddings, self.embedding_dim, self.padding_idx, self.max_norm, self.norm_type, self.scale_grad_by_freq, self.sparse,
                                      weights, self.data_bits, self.parameter_bits, self.training, self.momentum,
                                      running_x, running_w, running_o, scale_x, scale_w, scale_o,
                                      self.prefix, self.dump, self.path, self.quant_mode, self.o_bits, self.quant, self.is_not_from_iqtensor,
                                      self.clamp_data, self.clamp_weight, self.clamp_bias)
        self.running_x.fill_(running_x())
        self.running_w.fill_(running_w())
        self.running_o.fill_(running_o())
        self.scale_x.fill_(scale_x())
        self.scale_w.fill_(scale_w())
        self.scale_o.fill_(scale_o())
        return ret

    def state_dict(module, destination=None, prefix='', keep_vars=False):
        if destination is None:
            destination = OrderedDict()
            destination._metadata = OrderedDict()
        destination._metadata[prefix[:-1]
                              ] = local_metadata = dict(version=module._version)
        if is_in_onnx_export():
            assert module._buffers['running_w'] > 0, 'invalid running_w, please finetune first'

        if 'scale_w' in module._buffers and module._parameters['weight'].dtype == torch.float:
            weights = module._parameters['weight'].data
            q_weights, scale_mul_w, _ = module.quant.quant(
                weights, module.parameter_bits, mode=module.quant_mode)
            if is_in_onnx_export():
                scale_w = ScalerBuffer(scale_mul_w)
                module._buffers['scale_w'].data.fill_(scale_w())
                scale_o = scale_w
                module._buffers['scale_o'].data.fill_(scale_o())

            weight_tensor = module._parameters['weight']
            if is_in_onnx_export():

                if module.parameter_bits <= 8:
                    weight_tensor.data = q_weights.char()
                    weight_tensor.char()

                elif module.parameter_bits <= 16:
                    weight_tensor.data = q_weights.short()
                    weight_tensor.short()

                else:
                    weight_tensor.data = q_weights.int()
                    weight_tensor.int()

        module._save_to_state_dict(destination, prefix, keep_vars)
        for name, module in module._modules.items():
            if module is not None:
                module.state_dict(destination, prefix +
                                  name + '.', keep_vars=keep_vars)
        for hook in module._state_dict_hooks.values():
            hook_result = hook(module, destination, prefix, local_metadata)
            if hook_result is not None:
                destination = hook_result
        return destination

    def _load_from_state_dict(self, state_dict, prefix, local_metadata, strict,
                              missing_keys, unexpected_keys, error_msgs):
        ModuleIntConfig._load_from_state_dict_global(
            self, state_dict, prefix, local_metadata, strict, missing_keys, unexpected_keys, error_msgs)

    def extra_repr(self):
        s = nn.Embedding.extra_repr(self)
        extra_s = ',clamp_data:{clamp_data},clamp_weight:{clamp_weight},clamp_bias:{clamp_bias}'.format(
            **self.__dict__)
        extra_s += ',data_bits:{data_bits},parameter_bits:{parameter_bits},o_bits:{o_bits}'.format(
            **self.__dict__)
        return s+extra_s



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
from os import stat
import linger
import torch
import torch.nn as nn


def test_trace_iqtensor():

    class Add(nn.Module):
        def __init__(self) -> None:
            super().__init__()

        def forward(self, x, y):
            return x + y

    class iqtensorAdd(nn.Module):
        def __init__(self) -> None:
            super().__init__()
            self.add = Add()

        def forward(self, x, y):
            x = linger.quant_tensor(self, x, name='x')
            y = linger.quant_tensor(self, y, name='y')
            z = self.add(x, y)
            return z
    net = iqtensorAdd()
    dummy_input = torch.randn(10, 10)
    bb = torch.randn(10, 10)
    linger.trace_layers(net, net.add, (dummy_input, bb))
    cc = net(dummy_input, bb)
    with torch.no_grad():
        torch.onnx.export(net, (dummy_input, bb), "data.ignore/add.onnx", export_params=True, opset_version=12,
                          verbose=True, operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)


def test_iqview_onnx():
    class IQView(nn.Module):
        def __init__(self) -> None:
            super().__init__()
            self.conv = nn.Conv2d(2, 2, 2, 1, 1)
            self.fc = nn.Linear(18, 100)

        def forward(self, x):
            x = self.conv(x)
            n, c, h, w = x.shape
            x = x.view((n, c*h*w))
            x = self.fc(x)
            return x
    dummy_input = torch.randn(1, 2, 2, 2)
    net = IQView()
    out = net(dummy_input)

    net = linger.init(net)

    out = net(dummy_input)

    with torch.no_grad():
        torch.onnx.export(net, (dummy_input,), "data.ignore/iqview.onnx", export_params=True, opset_version=12,
                          verbose=True, operator_export_type=torch.onnx.OperatorExportTypes.ONNX_ATEN_FALLBACK)
